# image tests using pytest-mpl
from itertools import chain

import matplotlib as mpl
import numpy as np
import numpy.testing as npt
import pytest
from matplotlib.colors import SymLogNorm

from yt.data_objects.profiles import create_profile
from yt.loaders import load_uniform_grid
from yt.testing import add_noise_fields, fake_amr_ds, fake_particle_ds, fake_random_ds
from yt.visualization.api import (
    LinePlot,
    ParticleProjectionPlot,
    PhasePlot,
    ProfilePlot,
    SlicePlot,
)


def test_sliceplot_set_unit_and_zlim_order():
    ds = fake_random_ds(16)
    field = ("gas", "density")

    p0 = SlicePlot(ds, "z", field)
    p0.set_unit(field, "kg/m**3")
    p0.set_zlim(field, zmin=0)

    # reversing order of operations
    p1 = SlicePlot(ds, "z", field)
    p1.set_zlim(field, zmin=0)
    p1.set_unit(field, "kg/m**3")

    p0.render()
    p1.render()

    im0 = p0.plots[field].image._A
    im1 = p1.plots[field].image._A

    npt.assert_allclose(im0, im1)


def test_annotation_parse_h():
    ds = fake_random_ds(16)

    # Make sure `h` (reduced Hubble constant) is not equal to 1
    ds.unit_registry.modify("h", 0.7)

    rad = ds.quan(0.1, "cm/h")
    center = ds.arr([0.5] * 3, "code_length")

    # Twice the same slice plot
    p1 = SlicePlot(ds, "x", "density")
    p2 = SlicePlot(ds, "x", "density")

    # But the *same* center is given in different units
    p1.annotate_sphere(center.to("cm"), rad, circle_args={"color": "black"})
    p2.annotate_sphere(center.to("cm/h"), rad, circle_args={"color": "black"})

    # Render annotations, and extract matplotlib image
    # as an RGB array
    p1.render()
    p1.plots["gas", "density"].figure.canvas.draw()
    img1 = p1.plots["gas", "density"].figure.canvas.renderer.buffer_rgba()

    p2.render()
    p2.plots["gas", "density"].figure.canvas.draw()
    img2 = p2.plots["gas", "density"].figure.canvas.renderer.buffer_rgba()

    # This should be the same image
    npt.assert_allclose(img1, img2)


@pytest.mark.mpl_image_compare
def test_inf_and_finite_values_set_zlim():
    # see https://github.com/yt-project/yt/issues/3901
    shape = (32, 16, 1)
    a = np.ones(16)
    b = np.ones((32, 16))
    c = np.reshape(a * b, shape)

    # injecting an inf
    c[0, 0, 0] = np.inf

    field = ("gas", "density")
    data = {field: c}

    ds = load_uniform_grid(
        data,
        shape,
        bbox=np.array([[0, 1], [0, 1], [0, 1]]),
    )

    p = SlicePlot(ds, "z", field)

    # setting zlim manually
    p.set_zlim(field, -10, 10)

    p.render()
    return p.plots[field].figure


@pytest.mark.mpl_image_compare
def test_sliceplot_custom_norm():
    from matplotlib.colors import TwoSlopeNorm

    ds = fake_random_ds(16)
    field = ("gas", "density")
    p = SlicePlot(ds, "z", field)
    p.set_norm(field, norm=TwoSlopeNorm(vcenter=0, vmin=-0.5, vmax=1))

    p.render()
    return p.plots[field].figure


@pytest.mark.mpl_image_compare
def test_sliceplot_custom_norm_symlog_int_base():
    ds = fake_random_ds(16)
    add_noise_fields(ds)
    field = "noise3"
    p = SlicePlot(ds, "z", field)

    # using integer base !=10 and >2 to exercise special case
    # for colorbar minor ticks
    p.set_norm(field, norm=SymLogNorm(linthresh=0.1, base=5))

    p.render()
    return p.plots[field].figure


@pytest.mark.mpl_image_compare
def test_lineplot_set_axis_properties():
    ds = fake_random_ds(16)
    field = ("gas", "density")
    p = LinePlot(
        ds,
        field,
        start_point=[0, 0, 0],
        end_point=[1, 1, 1],
        npoints=32,
    )
    p.set_x_unit("cm")
    p.set_unit(field, "kg/cm**3")
    p.set_log(field, False)

    p.render()
    return p.plots[field].figure


@pytest.mark.mpl_image_compare
def test_profileplot_set_axis_properties():
    ds = fake_random_ds(16)

    disk = ds.disk(ds.domain_center, [0.0, 0.0, 1.0], (10, "m"), (1, "m"))
    p = ProfilePlot(disk, ("gas", "density"), [("gas", "velocity_x")])
    p.set_unit(("gas", "density"), "kg/cm**3")
    p.set_log(("gas", "density"), False)
    p.set_unit(("gas", "velocity_x"), "mile/hour")

    p.render()
    return p.plots["gas", "velocity_x"].figure


@pytest.mark.mpl_image_compare
def test_particleprojectionplot_set_colorbar_properties():
    ds = fake_particle_ds(npart=100)

    field = ("all", "particle_mass")
    p = ParticleProjectionPlot(ds, 2, field)
    p.set_buff_size(10)

    p.set_unit(field, "Msun")
    p.set_zlim(field, zmax=1e-35)
    p.set_log(field, False)

    p.render()
    return p.plots[field].figure


class TestMultipanelPlot:
    @classmethod
    def setup_class(cls):
        cls.fields = [
            ("gas", "density"),
            ("gas", "velocity_x"),
            ("gas", "velocity_y"),
            ("gas", "velocity_magnitude"),
        ]
        cls.ds = fake_random_ds(16)

    @pytest.mark.skipif(
        mpl.__version_info__ < (3, 7),
        reason="colorbar cannot currently be set horizontal in multi-panel plot with matplotlib older than 3.7.0",
    )
    @pytest.mark.parametrize("cbar_location", ["top", "bottom", "left", "right"])
    @pytest.mark.mpl_image_compare
    def test_multipanelplot_colorbar_orientation_simple(self, cbar_location):
        p = SlicePlot(self.ds, "z", self.fields)
        return p.export_to_mpl_figure((2, 2), cbar_location=cbar_location)

    @pytest.mark.parametrize("cbar_location", ["top", "bottom"])
    def test_multipanelplot_colorbar_orientation_warning(self, cbar_location):
        p = SlicePlot(self.ds, "z", self.fields)
        if mpl.__version_info__ < (3, 7):
            with pytest.warns(
                UserWarning,
                match="Cannot properly set the orientation of colorbar.",
            ):
                p.export_to_mpl_figure((2, 2), cbar_location=cbar_location)
        else:
            p.export_to_mpl_figure((2, 2), cbar_location=cbar_location)


class TestProfilePlot:
    @classmethod
    def setup_class(cls):
        fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
        units = ("g/cm**3", "K", "cm/s", "cm/s", "cm/s")
        cls.ds = fake_random_ds(16, fields=fields, units=units)
        regions = [cls.ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3), cls.ds.all_data()]
        pr_fields = [
            [("gas", "density"), ("gas", "temperature")],
            [("gas", "density"), ("gas", "velocity_x")],
            [("gas", "temperature"), ("gas", "mass")],
            [("gas", "density"), ("index", "radius")],
            [("gas", "velocity_magnitude"), ("gas", "mass")],
        ]
        cls.profiles: dict[str, ProfilePlot] = {}
        for i_reg, reg in enumerate(regions):
            id_prefix = str(i_reg)
            for x_field, y_field in pr_fields:
                id_suffix = "_".join([*x_field, *y_field])
                base_id = f"{id_prefix}_{id_suffix}"
                cls.profiles[base_id] = ProfilePlot(reg, x_field, y_field)
                cls.profiles[f"{base_id}_fractional_accumulation"] = ProfilePlot(
                    reg, x_field, y_field, fractional=True, accumulation=True
                )

                p1d = create_profile(reg, x_field, y_field)
                cls.profiles[f"{base_id}_from_profiles"] = ProfilePlot.from_profiles(
                    p1d
                )

        p1 = create_profile(
            cls.ds.all_data(), ("gas", "density"), ("gas", "temperature")
        )
        p2 = create_profile(
            cls.ds.all_data(), ("gas", "density"), ("gas", "velocity_x")
        )
        cls.profiles["from_multiple_profiles"] = ProfilePlot.from_profiles(
            [p1, p2], labels=["temperature", "velocity"]
        )

    @pytest.mark.parametrize(
        "suffix",
        [None, "from_profiles", "fractional_accumulation"],
    )
    @pytest.mark.parametrize("region", ["0", "1"])
    @pytest.mark.parametrize(
        "xax, yax",
        [
            (("gas", "density"), ("gas", "temperature")),
            (("gas", "density"), ("gas", "velocity_x")),
            (("gas", "temperature"), ("gas", "mass")),
            (("gas", "density"), ("index", "radius")),
            (("gas", "velocity_magnitude"), ("gas", "mass")),
        ],
    )
    @pytest.mark.mpl_image_compare
    def test_profileplot_simple(self, region, xax, yax, suffix):
        key = "_".join(chain(region, xax, yax))
        if suffix is not None:
            key += f"_{suffix}"
        plots = list(self.profiles[key].plots.values())
        assert len(plots) == 1
        return plots[0].figure

    @pytest.mark.mpl_image_compare
    def test_profileplot_from_multiple_profiles_0(self):
        plots = list(self.profiles["from_multiple_profiles"].plots.values())
        assert len(plots) == 2
        return plots[0].figure

    @pytest.mark.mpl_image_compare
    def test_profileplot_from_multiple_profiles_1(self):
        plots = list(self.profiles["from_multiple_profiles"].plots.values())
        assert len(plots) == 2
        return plots[1].figure


class TestPhasePlot:
    @classmethod
    def setup_class(cls):
        fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
        units = ("g/cm**3", "K", "cm/s", "cm/s", "cm/s")
        cls.ds = fake_random_ds(16, fields=fields, units=units)
        regions = [cls.ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3), cls.ds.all_data()]
        pr_fields = [
            [("gas", "density"), ("gas", "temperature"), ("gas", "mass")],
            [("gas", "density"), ("gas", "velocity_x"), ("gas", "mass")],
            [
                ("index", "radius"),
                ("gas", "temperature"),
                ("gas", "velocity_magnitude"),
            ],
        ]
        cls.profiles: dict[str, PhasePlot] = {}
        for i_reg, reg in enumerate(regions):
            id_prefix = str(i_reg)
            for x_field, y_field, z_field in pr_fields:
                id_suffix = "_".join([*x_field, *y_field, *z_field])
                base_id = f"{id_prefix}_{id_suffix}"
                cls.profiles[base_id] = PhasePlot(
                    reg, x_field, y_field, z_field, x_bins=16, y_bins=16
                )
                cls.profiles[f"{base_id}_fractional_accumulation"] = PhasePlot(
                    reg,
                    x_field,
                    y_field,
                    z_field,
                    fractional=True,
                    accumulation=True,
                    x_bins=16,
                    y_bins=16,
                )

                p2d = create_profile(reg, [x_field, y_field], z_field, n_bins=[16, 16])
                cls.profiles[f"{base_id}_from_profiles"] = PhasePlot.from_profile(p2d)

    @pytest.mark.parametrize(
        "suffix",
        [None, "from_profiles", "fractional_accumulation"],
    )
    @pytest.mark.parametrize("region", ["0", "1"])
    @pytest.mark.parametrize(
        "xax, yax, zax",
        [
            (("gas", "density"), ("gas", "temperature"), ("gas", "mass")),
            (("gas", "density"), ("gas", "velocity_x"), ("gas", "mass")),
            (
                ("index", "radius"),
                ("gas", "temperature"),
                ("gas", "velocity_magnitude"),
            ),
        ],
    )
    @pytest.mark.mpl_image_compare
    def test_phaseplot(self, region, xax, yax, zax, suffix):
        key = "_".join(chain(region, xax, yax, zax))
        if suffix is not None:
            key += f"_{suffix}"
        plots = list(self.profiles[key].plots.values())
        assert len(plots) == 1
        return plots[0].figure


class TestPhasePlotSetZlim:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_random_ds(16)
        add_noise_fields(cls.ds)
        cls.data = cls.ds.sphere("c", 1)

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_zlim_with_implicit_units(self):
        p = PhasePlot(
            self.data,
            ("gas", "noise1"),
            ("gas", "noise3"),
            [("gas", "density")],
            weight_field=None,
        )
        field = ("gas", "density")
        p.set_zlim(field, zmax=10)
        p.render()
        return p.plots[field].figure

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_zlim_with_explicit_units(self):
        p = PhasePlot(
            self.data,
            ("gas", "noise1"),
            ("gas", "noise3"),
            [("gas", "density")],
            weight_field=None,
        )
        field = ("gas", "density")
        # using explicit units, we expect the colorbar units to stay unchanged
        p.set_zlim(field, zmin=(1e36, "kg/AU**3"))
        p.render()
        return p.plots[field].figure


class TestSetBackgroundColor:
    # see https://github.com/yt-project/yt/issues/3854

    @classmethod
    def setup_class(cls):
        cls.ds = fake_random_ds(16)

        def some_nans_field(field, data):
            ret = data["gas", "density"]
            ret[::2] *= np.nan
            return ret

        cls.ds.add_field(
            name=("gas", "polluted_field"),
            function=some_nans_field,
            sampling_type="local",
        )

    @pytest.mark.mpl_image_compare
    def test_sliceplot_set_background_color_linear(self):
        field = ("gas", "density")
        p = SlicePlot(self.ds, "z", field, width=1.5)
        p.set_background_color(field, color="C0")
        p.set_log(field, False)

        p.render()
        return p.plots[field].figure

    @pytest.mark.mpl_image_compare
    def test_sliceplot_set_background_color_log(self):
        field = ("gas", "density")
        p = SlicePlot(self.ds, "z", field, width=1.5)
        p.set_background_color(field, color="C0")

        p.render()
        return p.plots[field].figure

    @pytest.mark.mpl_image_compare
    def test_sliceplot_set_background_color_and_bad_value(self):
        # see https://github.com/yt-project/yt/issues/4639
        field = ("gas", "polluted_field")
        p = SlicePlot(self.ds, "z", field, width=1.5)
        p.set_background_color(field, color="black")

        # copy the default colormap
        cmap = mpl.colormaps["cmyt.arbre"].with_extremes(bad="red")
        p.set_cmap(field, cmap)

        p.render()
        return p.plots[field].figure


class TestCylindricalZSlicePlot:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_amr_ds(geometry="cylindrical")
        add_noise_fields(cls.ds)
        fields = ["noise%d" % i for i in range(4)]
        cls.plot = SlicePlot(cls.ds, "z", fields)

    @pytest.mark.parametrize("field", ["noise0", "noise1", "noise2", "noise3"])
    @pytest.mark.mpl_image_compare
    def test_cylindrical_z_log(self, field):
        return self.plot.plots[field].figure

    @pytest.mark.parametrize("field", ["noise0", "noise1", "noise2", "noise3"])
    @pytest.mark.mpl_image_compare
    def test_cylindrical_z_linear(self, field):
        self.plot.set_log("noise0", False)
        return self.plot.plots[field].figure

    @pytest.mark.parametrize(
        "theta_min, theta_max",
        [
            pytest.param(0, 2 * np.pi, id="full_azimuthal_domain"),
            pytest.param(3 / 4 * np.pi, 5 / 4 * np.pi, id="restricted_sector"),
        ],
    )
    @pytest.mark.mpl_image_compare
    def test_exclude_pixels_with_partial_bbox_intersection(self, theta_min, theta_max):
        rmin = 1.0
        rmax = 2.0
        ds = fake_amr_ds(
            geometry="cylindrical",
            domain_left_edge=[rmin, 0, theta_min],
            domain_right_edge=[rmax, 1, theta_max],
        )
        add_noise_fields(ds)
        plot = SlicePlot(ds, "z", ("gas", "noise0"))
        for radius in [rmin - 0.01, rmax]:
            plot.annotate_sphere(
                center=[0, 0, 0],
                radius=radius,
                circle_args={"color": "red", "alpha": 0.4, "linewidth": 3},
            )
        plot.annotate_title("all pixels beyond (or on) red lines should be white")
        plot.render()
        return plot.plots["gas", "noise0"].figure


class TestSphericalPhiSlicePlot:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_amr_ds(geometry="spherical")
        add_noise_fields(cls.ds)
        fields = ["noise%d" % i for i in range(4)]
        cls.plot = SlicePlot(cls.ds, "phi", fields)

    @pytest.mark.parametrize("field", ["noise0", "noise1", "noise2", "noise3"])
    @pytest.mark.mpl_image_compare
    def test_spherical_phi_log(self, field):
        return self.plot.plots[field].figure


class TestSphericalThetaSlicePlot:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_amr_ds(geometry="spherical")
        add_noise_fields(cls.ds)
        fields = ["noise%d" % i for i in range(4)]
        cls.plot = SlicePlot(cls.ds, "theta", fields)

    @pytest.mark.parametrize("field", ["noise0", "noise1", "noise2", "noise3"])
    @pytest.mark.mpl_image_compare
    def test_spherical_theta_log(self, field):
        return self.plot.plots[field].figure



import numpy as np

import yt
from yt.testing import requires_file, requires_module
from yt.utilities.on_demand_imports import _h5py as h5py


@requires_file("snapshot_033/snap_033.0.hdf5")
@requires_module("h5py")
def test_gadget_header_array_reduction(tmp_path):
    # first get a real header
    ds = yt.load("snapshot_033/snap_033.0.hdf5")
    hvals = ds._get_hvals()
    hvals_orig = hvals.copy()
    # wrap some of the scalar values in nested arrays
    hvals["Redshift"] = np.array([hvals["Redshift"]])
    hvals["Omega0"] = np.array([[hvals["Omega0"]]])

    # drop those header values into a fake header-only file
    tmp_snpshot_dir = tmp_path / "snapshot_033"
    tmp_snpshot_dir.mkdir()
    tmp_header_only_file = str(tmp_snpshot_dir / "fake_gadget_header.hdf5")
    with h5py.File(tmp_header_only_file, mode="w") as f:
        headergrp = f.create_group("Header")
        for field, val in hvals.items():
            headergrp.attrs[field] = val

    # trick the dataset into using the header file and make sure the
    # arrays are reduced
    ds._input_filename = tmp_header_only_file
    hvals = ds._get_hvals()
    for attr in ("Redshift", "Omega0"):
        assert hvals[attr] == hvals_orig[attr]
        assert isinstance(hvals[attr], np.ndarray) is False



from yt.fields.field_info_container import FieldInfoContainer

# We need to specify which fields we might have in our dataset.  The field info
# container subclass here will define which fields it knows about.  There are
# optionally methods on it that get called which can be subclassed.


class SkeletonFieldInfo(FieldInfoContainer):
    known_other_fields = (
        # Each entry here is of the form
        # ( "name", ("units", ["fields", "to", "alias"], # "display_name")),
    )

    known_particle_fields = (
        # Identical form to above
        # ( "name", ("units", ["fields", "to", "alias"], # "display_name")),
    )

    def __init__(self, ds, field_list):
        super().__init__(ds, field_list)
        # If you want, you can check self.field_list

    def setup_fluid_fields(self):
        # Here we do anything that might need info about the dataset.
        # You can use self.alias, self.add_output_field (for on-disk fields)
        # and self.add_field (for derived fields).
        pass

    def setup_particle_fields(self, ptype):
        super().setup_particle_fields(ptype)
        # This will get called for every particle type.



import cmd
import signal
import sys
import traceback
from io import StringIO
from xmlrpc.client import ServerProxy
from xmlrpc.server import SimpleXMLRPCServer

from yt.config import ytcfg


class PdbXMLRPCServer(SimpleXMLRPCServer):
    """
    shutdown-enabled XMLRPCServer from
      http://code.activestate.com/recipes/114579/
    """

    finished = False

    def register_signal(self, signum):
        signal.signal(signum, self.signal_handler)

    def signal_handler(self, signum, frame):
        print("Caught signal", signum)
        self.shutdown()

    def shutdown(self):
        self.finished = True
        return 1

    def serve_forever(self):
        while not self.finished:
            self.handle_request()
        print("DONE SERVING")


def rpdb_excepthook(exc_type, exc, tb):
    traceback.print_exception(exc_type, exc, tb)
    task = ytcfg.get("yt", "internals", "global_parallel_rank")
    size = ytcfg.get("yt", "internals", "global_parallel_size")
    print(f"Starting RPDB server on task {task} ; connect with 'yt rpdb -t {task}'")
    handler = pdb_handler(tb)
    server = PdbXMLRPCServer(("localhost", 8010 + task))
    server.register_introspection_functions()
    server.register_instance(handler)
    server.register_function(server.shutdown)
    server.serve_forever()
    server.server_close()
    if size > 1:
        from mpi4py import MPI

        # This COMM_WORLD is okay.  We want to barrierize here, while waiting
        # for shutdown from the rest of the parallel group.  If you are running
        # with --rpdb it is assumed you know what you are doing and you won't
        # let this get out of hand.
        MPI.COMM_WORLD.Barrier()


class pdb_handler:
    def __init__(self, tb):
        import pdb

        self.cin = StringIO()
        sys.stdin = self.cin
        self.cout = StringIO()
        sys.stdout = self.cout
        sys.stderr = self.cout
        self.debugger = pdb.Pdb(stdin=self.cin, stdout=self.cout)
        self.debugger.reset()
        self.debugger.setup(tb.tb_frame, tb)

    def execute(self, line):
        tt = self.cout.tell()
        self.debugger.onecmd(line)
        self.cout.seek(tt)
        return self.cout.read()


class rpdb_cmd(cmd.Cmd):
    def __init__(self, proxy):
        self.proxy = proxy
        cmd.Cmd.__init__(self)
        print(self.proxy.execute("bt"))

    def default(self, line):
        print(self.proxy.execute(line))

    def do_shutdown(self, args):
        print(self.proxy.shutdown())
        return True

    def do_help(self, line):
        print(self.proxy.execute(f"help {line}"))

    def postcmd(self, stop, line):
        return stop

    def postloop(self):
        try:
            self.proxy.shutdown()
        except Exception:
            pass


__header = """
You're in a remote PDB session with task %(task)s

You can run PDB commands, and when you're done, type 'shutdown' to quit.
"""


def run_rpdb(task=None):
    port = 8010
    if task is None:
        try:
            task + int(sys.argv[-1])
        except Exception:
            pass
    port += task
    sp = ServerProxy(f"http://localhost:{port}/")
    try:
        pp = rpdb_cmd(sp)
    except OSError:
        print("Connection refused.  Is the server running?")
        sys.exit(1)
    pp.cmdloop(__header % {"task": port - 8010})



import pytest

from yt.frontends.amrex.fields import Substance


@pytest.mark.parametrize(
    "data, expected",
    [
        pytest.param("X(He5)", [("He", 5)], id="isotope_1"),
        pytest.param("X(C12)", [("C", 12)], id="isotope_2"),
        pytest.param("X(A1B2C3)", [("A", 1), ("B", 2), ("C", 3)], id="molecule_1"),
        pytest.param("X(C12H24)", [("C", 12), ("H", 24)], id="molecule_2"),
        pytest.param("X(H2O)", [("H", 2), ("O", 1)], id="molecule_3"),
        pytest.param("X(ash)", [("ash", 0)], id="descriptive_name"),
    ],
)
def test_Substance_spec(data, expected):
    assert Substance(data)._spec == expected


@pytest.mark.parametrize(
    "data, expected_type",
    [
        pytest.param("X(He5)", "isotope", id="isotope_1"),
        pytest.param("X(C12)", "isotope", id="isotope_2"),
        pytest.param("X(A1B2C3)", "molecule", id="molecule_1"),
        pytest.param("X(C12H24)", "molecule", id="molecule_2"),
        pytest.param("X(H2O)", "molecule", id="molecule_3"),
        pytest.param("X(ash)", "descriptive_name", id="descriptive_name"),
    ],
)
def test_Substance_type(data, expected_type):
    sub = Substance(data)
    assert getattr(sub, f"is_{expected_type}")()


@pytest.mark.parametrize(
    "data, expected_str, expected_tex",
    [
        pytest.param("X(He5)", "He5", "^{5}He", id="isotope_1"),
        pytest.param("X(C12)", "C12", "^{12}C", id="isotope_2"),
        pytest.param("X(A1B2C3)", "AB2C3", "A_{}B_{2}C_{3}", id="molecule_1"),
        pytest.param("X(C12H24)", "C12H24", "C_{12}H_{24}", id="molecule_2"),
        pytest.param("X(H2O)", "H2O", "H_{2}O_{}", id="molecule_2"),
        pytest.param("X(ash)", "ash", "ash", id="descriptive_name"),
    ],
)
def test_Substance_to_str(data, expected_str, expected_tex):
    sub = Substance(data)
    assert str(sub) == expected_str
    assert sub.to_tex() == expected_tex



import numpy as np
from numpy.testing import assert_equal

from yt.testing import fake_random_ds


def test_ortho_ray():
    ds = fake_random_ds(64, nprocs=8)
    dx = (ds.domain_right_edge - ds.domain_left_edge) / ds.domain_dimensions

    axes = ["x", "y", "z"]
    for ax in range(3):
        ocoord = ds.arr(np.random.random(2), "code_length")

        my_oray = ds.ortho_ray(ax, ocoord)

        my_axes = ds.coordinates.x_axis[ax], ds.coordinates.y_axis[ax]

        # find the cells intersected by the ortho ray
        my_all = ds.all_data()
        my_cells = (
            np.abs(my_all["index", axes[my_axes[0]]] - ocoord[0])
            <= 0.5 * dx[my_axes[0]]
        ) & (
            np.abs(my_all["index", axes[my_axes[1]]] - ocoord[1])
            <= 0.5 * dx[my_axes[1]]
        )

        assert_equal(
            my_oray["gas", "density"].sum(),
            my_all["gas", "density"][my_cells].sum(),
        )



import weakref
from functools import cached_property

from yt.funcs import obj_length
from yt.units.yt_array import YTQuantity
from yt.utilities.exceptions import (
    YTDimensionalityError,
    YTFieldNotFound,
    YTFieldNotParseable,
)
from yt.visualization.line_plot import LineBuffer

from .data_containers import _get_ipython_key_completion


class RegionExpression:
    def __init__(self, ds):
        self.ds = weakref.proxy(ds)

    @cached_property
    def all_data(self):
        return self.ds.all_data()

    def __getitem__(self, item):
        # At first, we will only implement this as accepting a slice that is
        # (optionally) unitful corresponding to a specific set of coordinates
        # that result in a rectangular prism or a slice.
        try:
            return self.all_data[item]
        except (YTFieldNotParseable, YTFieldNotFound):
            # any error raised by self.ds._get_field_info
            # signals a type error (not a field), however we don't want to
            # catch plain TypeErrors as this may create subtle bugs very hard
            # to decipher, like broken internal function calls.
            pass

        if isinstance(item, slice):
            if obj_length(item.start) == 3 and obj_length(item.stop) == 3:
                # This is for a ray that is not orthogonal to an axis.
                # it's straightforward to do this, so we create a ray
                # and drop out here.
                return self._create_ray(item)
            else:
                # This is for the case where we give a slice as an index; one
                # possible use case of this would be where we supply something
                # like ds.r[::256j] .  This would be expanded, implicitly into
                # ds.r[::256j, ::256j, ::256j].  Other cases would be if we do
                # ds.r[0.1:0.9] where it will be expanded along all dimensions.
                item = tuple(item for _ in range(self.ds.dimensionality))

        if item is Ellipsis:
            item = (Ellipsis,)

        # from this point, item is implicitly assumed to be iterable
        if Ellipsis in item:
            # expand "..." into the appropriate number of ":"
            item = list(item)
            idx = item.index(Ellipsis)
            item.pop(idx)
            if Ellipsis in item:
                # this error mimics numpy's
                raise IndexError("an index can only have a single ellipsis ('...')")
            while len(item) < self.ds.dimensionality:
                item.insert(idx, slice(None))

        if len(item) != self.ds.dimensionality:
            # Not the right specification, and we don't want to do anything
            # implicitly.  Note that this happens *after* the implicit expansion
            # of a single slice.
            raise YTDimensionalityError(len(item), self.ds.dimensionality)

        # OK, now we need to look at our slices.  How many are a specific
        # coordinate?

        nslices = sum(isinstance(v, slice) for v in item)
        if nslices == 0:
            return self._create_point(item)
        elif nslices == 1:
            return self._create_ortho_ray(item)
        elif nslices == 2:
            return self._create_slice(item)
        else:
            if all(s.start is s.stop is s.step is None for s in item):
                return self.all_data
            return self._create_region(item)

    def _ipython_key_completions_(self):
        return _get_ipython_key_completion(self.ds)

    def _spec_to_value(self, input):
        if isinstance(input, tuple):
            v = self.ds.quan(input[0], input[1]).to("code_length")
        elif isinstance(input, YTQuantity):
            v = self.ds.quan(input).to("code_length")
        else:
            v = self.ds.quan(input, "code_length")
        return v

    def _create_slice(self, slice_tuple):
        # This is somewhat more complex because we want to allow for slicing
        # in one dimension but also *not* using the entire domain; for instance
        # this means we allow something like ds.r[0.5, 0.1:0.4, 0.1:0.4].
        axis = None
        new_slice = []
        for ax, v in enumerate(slice_tuple):
            if not isinstance(v, slice):
                if axis is not None:
                    raise RuntimeError
                axis = ax
                coord = self._spec_to_value(v)
                new_slice.append(slice(None, None, None))
            else:
                new_slice.append(v)
        # This new slice doesn't need to be a tuple
        dim = self.ds.dimensionality
        if dim < 2:
            raise ValueError(
                "Can not create a slice from data with dimensionality '%d'" % dim
            )
        if dim == 2:
            coord = self.ds.domain_center[2]
            axis = 2
        source = self._create_region(new_slice)
        sl = self.ds.slice(axis, coord, data_source=source)
        # Now, there's the possibility that what we're also seeing here
        # includes some steps, which would be for getting back a fixed
        # resolution buffer.  We check for that by checking if we have
        # exactly two imaginary steps.
        xax = self.ds.coordinates.x_axis[axis]
        yax = self.ds.coordinates.y_axis[axis]
        if (
            getattr(new_slice[xax].step, "imag", 0.0) != 0.0
            and getattr(new_slice[yax].step, "imag", 0.0) != 0.0
        ):
            # We now need to convert to a fixed res buffer.
            # We'll do this by getting the x/y axes, and then using that.
            width = source.right_edge[xax] - source.left_edge[xax]
            height = source.right_edge[yax] - source.left_edge[yax]
            # Make a resolution tuple with
            resolution = (int(new_slice[xax].step.imag), int(new_slice[yax].step.imag))
            # Use the center of the slice, not the entire domain
            center = source.center
            sl = sl.to_frb(
                width=width, resolution=resolution, height=height, center=center
            )
        return sl

    def _slice_to_edges(self, ax, val):
        if val.start is None:
            l = self.ds.domain_left_edge[ax]
        else:
            l = self._spec_to_value(val.start)
        if val.stop is None:
            r = self.ds.domain_right_edge[ax]
        else:
            r = self._spec_to_value(val.stop)
        if r < l:
            raise RuntimeError
        return l, r

    def _create_region(self, bounds_tuple):
        left_edge = self.ds.domain_left_edge.copy()
        right_edge = self.ds.domain_right_edge.copy()
        dims = [1, 1, 1]
        for ax, b in enumerate(bounds_tuple):
            l, r = self._slice_to_edges(ax, b)
            left_edge[ax] = l
            right_edge[ax] = r
            d = getattr(b.step, "imag", None)
            if d is not None:
                d = int(d)
            dims[ax] = d
        center = (left_edge + right_edge) / 2.0
        if None not in dims:
            return self.ds.arbitrary_grid(left_edge, right_edge, dims)
        return self.ds.region(center, left_edge, right_edge)

    def _create_point(self, point_tuple):
        coord = [self._spec_to_value(p) for p in point_tuple]
        return self.ds.point(coord)

    def _create_ray(self, ray_slice):
        start_point = [self._spec_to_value(v) for v in ray_slice.start]
        end_point = [self._spec_to_value(v) for v in ray_slice.stop]
        if getattr(ray_slice.step, "imag", 0.0) != 0.0:
            return LineBuffer(self.ds, start_point, end_point, int(ray_slice.step.imag))
        else:
            return self.ds.ray(start_point, end_point)

    def _create_ortho_ray(self, ray_tuple):
        axis = None
        new_slice = []
        coord = []
        npoints = 0
        start_point = []
        end_point = []
        for ax, v in enumerate(ray_tuple):
            if not isinstance(v, slice):
                val = self._spec_to_value(v)
                coord.append(val)
                new_slice.append(slice(None, None, None))
                start_point.append(val)
                end_point.append(val)
            else:
                if axis is not None:
                    raise RuntimeError
                if getattr(v.step, "imag", 0.0) != 0.0:
                    npoints = int(v.step.imag)
                    xi = self._spec_to_value(v.start)
                    xf = self._spec_to_value(v.stop)
                    dx = (xf - xi) / npoints
                    start_point.append(xi + 0.5 * dx)
                    end_point.append(xf - 0.5 * dx)
                else:
                    axis = ax
                    new_slice.append(v)
        if npoints > 0:
            ray = LineBuffer(self.ds, start_point, end_point, npoints)
        else:
            if axis == 1:
                coord = [coord[1], coord[0]]
            source = self._create_region(new_slice)
            ray = self.ds.ortho_ray(axis, coord, data_source=source)
        return ray



import numpy as np

# Now define convenience functions


def blankRecordArray(desc, elements):
    """
    Accept a descriptor describing a recordarray, and return one that's full of
    zeros

    This seems like it should be in the numpy distribution...
    """
    blanks = []
    for atype in desc["formats"]:
        blanks.append(np.zeros(elements, dtype=atype))
    return np.rec.fromarrays(blanks, **desc)



import os.path
from itertools import islice

from yt.config import ytcfg
from yt.funcs import mylog
from yt.utilities.object_registries import output_type_registry
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    parallel_simple_proxy,
)

_field_names = ("hash", "bn", "fp", "tt", "ctid", "class_name", "last_seen")


class NoParameterShelf(Exception):
    pass


class UnknownDatasetType(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"{self.name}"


class ParameterFileStore:
    """
    This class is designed to be a semi-persistent storage for parameter
    files.  By identifying each dataset with a unique hash, objects
    can be stored independently of datasets -- when an object is
    loaded, the dataset is as well, based on the hash.  For
    storage concerns, only a few hundred will be retained in cache.

    """

    _shared_state = {}  # type: ignore
    _distributed = True
    _processing = False
    _owner = 0
    _register = True

    def __new__(cls, *p, **k):
        self = object.__new__(cls, *p, **k)
        self.__dict__ = cls._shared_state
        return self

    def __init__(self, in_memory=False):
        """
        Create the dataset database if yt is configured to store them.
        Otherwise, use read-only settings.

        """
        if not self._register:
            return
        if ytcfg.get("yt", "store_parameter_files"):
            self._read_only = False
            self.init_db()
            self._records = self.read_db()
        else:
            self._read_only = True
            self._records = {}
        self._register = False

    @parallel_simple_proxy
    def init_db(self):
        """
        This function ensures that the storage database exists and can be used.
        """
        dbn = self._get_db_name()
        dbdir = os.path.dirname(dbn)
        try:
            if not os.path.isdir(dbdir):
                os.mkdir(dbdir)
        except OSError as exc:
            raise NoParameterShelf from exc
        open(dbn, "ab")  # make sure it exists, allow to close
        # Now we read in all our records and return them
        # these will be broadcast

    def _get_db_name(self):
        base_file_name = ytcfg.get("yt", "parameter_file_store")
        if not os.access(os.path.expanduser("~/"), os.W_OK):
            return os.path.abspath(base_file_name)
        return os.path.expanduser(f"~/.yt/{base_file_name}")

    def get_ds_hash(self, hash):
        """This returns a dataset based on a hash."""
        return self._convert_ds(self._records[hash])

    def get_ds_ctid(self, ctid):
        """This returns a dataset based on a CurrentTimeIdentifier."""
        for h in self._records:
            if self._records[h]["ctid"] == ctid:
                return self._convert_ds(self._records[h])

    def _adapt_ds(self, ds):
        """This turns a dataset into a CSV entry."""
        return {
            "bn": ds.basename,
            "fp": ds.directory,
            "tt": ds.current_time,
            "ctid": ds.unique_identifier,
            "class_name": ds.__class__.__name__,
            "last_seen": ds._instantiated,
        }

    def _convert_ds(self, ds_dict):
        """This turns a CSV entry into a dataset."""
        bn = ds_dict["bn"]
        fp = ds_dict["fp"]
        fn = os.path.join(fp, bn)
        class_name = ds_dict["class_name"]
        if class_name not in output_type_registry:
            raise UnknownDatasetType(class_name)
        mylog.info("Checking %s", fn)
        if os.path.exists(fn):
            ds = output_type_registry[class_name](os.path.join(fp, bn))
        else:
            raise OSError
        # This next one is to ensure that we manually update the last_seen
        # record *now*, for during write_out.
        self._records[ds._hash()]["last_seen"] = ds._instantiated
        return ds

    def check_ds(self, ds):
        """
        This will ensure that the dataset (*ds*) handed to it is
        recorded in the storage unit.  In doing so, it will update path
        and "last_seen" information.
        """
        hash = ds._hash()
        if hash not in self._records:
            self.insert_ds(ds)
            return
        ds_dict = self._records[hash]
        self._records[hash]["last_seen"] = ds._instantiated
        if ds_dict["bn"] != ds.basename or ds_dict["fp"] != ds.directory:
            self.wipe_hash(hash)
            self.insert_ds(ds)

    def insert_ds(self, ds):
        """This will insert a new *ds* and flush the database to disk."""
        self._records[ds._hash()] = self._adapt_ds(ds)
        self.flush_db()

    def wipe_hash(self, hash):
        """
        This removes a *hash* corresponding to a dataset from the
        storage.
        """
        if hash not in self._records:
            return
        del self._records[hash]
        self.flush_db()

    def flush_db(self):
        """This flushes the storage to disk."""
        if self._read_only:
            return
        self._write_out()
        self.read_db()

    def get_recent(self, n=10):
        recs = sorted(self._records.values(), key=lambda a: -a["last_seen"])[:n]
        return recs

    @parallel_simple_proxy
    def _write_out(self):
        import csv

        if self._read_only:
            return
        fn = self._get_db_name()
        f = open(f"{fn}.tmp", "w")
        w = csv.DictWriter(f, _field_names)
        maxn = ytcfg.get("yt", "maximum_stored_datasets")  # number written
        for h, v in islice(
            sorted(self._records.items(), key=lambda a: -a[1]["last_seen"]), 0, maxn
        ):
            v["hash"] = h
            w.writerow(v)
        f.close()
        os.rename(f"{fn}.tmp", fn)

    @parallel_simple_proxy
    def read_db(self):
        """This will read the storage device from disk."""
        import csv

        f = open(self._get_db_name())
        vals = csv.DictReader(f, _field_names)
        db = {}
        for v in vals:
            db[v.pop("hash")] = v
            if v["last_seen"] is None:
                v["last_seen"] = 0.0
            else:
                v["last_seen"] = float(v["last_seen"])
        f.close()
        return db



# This extension is quite simple:
#  1. It accepts a script name
#  2. This script is added to the document in a literalinclude
#  3. Any _static images found will be added

import glob
import os
import shutil

from docutils.parsers.rst import Directive, directives

# Some of this magic comes from the matplotlib plot_directive.


def setup(app):
    app.add_directive("yt_cookbook", CookbookScript)
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    retdict = dict(version="0.1", parallel_read_safe=True, parallel_write_safe=True)

    return retdict


data_patterns = ["*.h5", "*.out", "*.dat", "*.mp4"]


class CookbookScript(Directive):
    required_arguments = 1
    optional_arguments = 0

    def run(self):
        rst_file = self.state_machine.document.attributes["source"]
        rst_dir = os.path.abspath(os.path.dirname(rst_file))
        script_fn = directives.path(self.arguments[0])
        script_bn = os.path.basename(script_fn)
        script_name = os.path.basename(self.arguments[0]).split(".")[0]

        # This magic is from matplotlib
        dest_dir = os.path.abspath(
            os.path.join(setup.app.builder.outdir, os.path.dirname(script_fn))
        )
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)  # no problem here for me, but just use built-ins

        rel_dir = os.path.relpath(rst_dir, setup.confdir)
        place = os.path.join(dest_dir, rel_dir)
        if not os.path.isdir(place):
            os.makedirs(place)
        shutil.copyfile(
            os.path.join(rst_dir, script_fn), os.path.join(place, script_bn)
        )

        im_path = os.path.join(rst_dir, "_static")
        images = sorted(glob.glob(os.path.join(im_path, f"{script_name}__*.png")))
        lines = []
        lines.append(f"(`{script_bn} <{script_fn}>`__)")
        lines.append("\n")
        lines.append("\n")
        lines.append(f".. literalinclude:: {self.arguments[0]}")
        lines.append("\n")
        lines.append("\n")
        for im in images:
            im_name = os.path.join("_static", os.path.basename(im))
            lines.append(f".. image:: {im_name}")
            lines.append("   :width: 400")
            lines.append(f"   :target: ../_images/{os.path.basename(im)}")
            lines.append("\n")
        lines.append("\n")
        for ext in data_patterns:
            data_files = sorted(
                glob.glob(os.path.join(im_path, f"{script_name}__*.{ext}"))
            )
            for df in data_files:
                df_bn = os.path.basename(df)
                shutil.copyfile(
                    os.path.join(rst_dir, df), os.path.join(dest_dir, rel_dir, df_bn)
                )
                lines.append(f" * Data: `{df_bn} <{df}>`__)")
            lines.append("\n")
        self.state_machine.insert_input(lines, rst_file)
        return []



import numpy as np
from numpy.testing import assert_array_equal

from yt.utilities.lib.bounded_priority_queue import (
    validate,
    validate_nblist,
    validate_pid,
)


# These test functions use utility functions in
# yt.utilities.lib.bounded_priority_queue
# to test functions which are not exposed at a python level
def test_bounded_priority_queue():
    dists = validate()
    answers = np.array([0.1, 0.001, -1.0, -1.0, -1.0])
    assert_array_equal(answers, dists)


def test_bounded_priority_queue_pid():
    dists, pids = validate_pid()
    answers = np.array([0.1, 0.001, -1.0, -1.0, -1.0])
    answers_pids = np.array([1, 10, -1, -1, -1])
    assert_array_equal(answers, dists)
    assert_array_equal(answers_pids, pids)


def test_neighbor_list():
    data, pids = validate_nblist()
    answers_data = np.array([1.0, 1.0, 1.0, 1.0])
    answers_pids = np.array([0, 1, 2, 3])
    assert_array_equal(answers_data, data)
    assert_array_equal(answers_pids, pids)




# distutils: libraries = STD_LIBS
"""
AMR kD-Tree Cython Tools



"""


import numpy as np

cimport cython
cimport numpy as np
from cython.view cimport array as cvarray
from libc.stdlib cimport free, malloc

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef class Node:

    def __cinit__(self,
                  Node parent,
                  Node left,
                  Node right,
                  np.float64_t[:] left_edge,
                  np.float64_t[:] right_edge,
                  int grid,
                  np.int64_t node_id):
        self.dirty = False
        self.left = left
        self.right = right
        self.parent = parent
        cdef int i
        for i in range(3):
            self.left_edge[i] = left_edge[i]
            self.right_edge[i] = right_edge[i]
        self.grid = grid
        self.node_id = node_id
        self.split == NULL


    def print_me(self):
        print('Node %i' % self.node_id)
        print('\t le: %e %e %e' % (self.left_edge[0], self.left_edge[1],
                                   self.left_edge[2]))
        print('\t re: %e %e %e' % (self.right_edge[0], self.right_edge[1],
                                   self.right_edge[2]))
        print('\t grid: %i' % self.grid)

    def get_split_dim(self):
        if self.split != NULL:
            return self.split.dim
        else:
            return -1

    def get_split_pos(self):
        if self.split != NULL:
            return self.split.pos
        else:
            return np.nan

    def set_left_edge(self, np.float64_t[:] left_edge):
        cdef int i
        for i in range(3):
            self.left_edge[i] = left_edge[i]

    def set_right_edge(self, np.float64_t[:] right_edge):
        cdef int i
        for i in range(3):
            self.right_edge[i] = right_edge[i]

    def create_split(self, dim, pos):
        split = <Split *> malloc(sizeof(Split))
        split.dim = dim
        split.pos = pos
        self.split = split

    def __dealloc__(self):
        if self.split != NULL: free(self.split)

    # Begin input of converted methods

    def get_left_edge(self):
        le = np.empty(3, dtype='float64')
        for i in range(3):
            le[i] = self.left_edge[i]
        return le

    def get_right_edge(self):
        re = np.empty(3, dtype='float64')
        for i in range(3):
            re[i] = self.right_edge[i]
        return re

    def set_dirty(self, bint state):
        cdef Node node
        for node in self.depth_traverse():
            node.dirty = state

    def kd_traverse(self, viewpoint=None):
        cdef Node node
        if viewpoint is None:
            for node in self.depth_traverse():
                if node._kd_is_leaf() == 1 and node.grid != -1:
                    yield node
        else:
            for node in self.viewpoint_traverse(viewpoint):
                if node._kd_is_leaf() == 1 and node.grid != -1:
                    yield node

    def add_pygrid(self,
                       np.float64_t[:] gle,
                       np.float64_t[:] gre,
                       int gid,
                       int rank,
                       int size):

        """
        The entire purpose of this function is to move everything from ndarrays
        to internal C pointers.
        """
        cdef np.float64_t[:] pgles = cvarray(format="d", shape=(3,), itemsize=sizeof(np.float64_t))
        cdef np.float64_t[:] pgres = cvarray(format="d", shape=(3,), itemsize=sizeof(np.float64_t))
        cdef int j
        for j in range(3):
            pgles[j] = gle[j]
            pgres[j] = gre[j]

        self.add_grid(pgles, pgres, gid, rank, size)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef add_grid(self,
                       np.float64_t[:] gle,
                       np.float64_t[:] gre,
                       int gid,
                       int rank,
                       int size):

        if not should_i_build(self, rank, size):
            return

        if self._kd_is_leaf() == 1:
            self.insert_grid(gle, gre, gid, rank, size)
        else:
            less_id = gle[self.split.dim] < self.split.pos
            if less_id:
                self.left.add_grid(gle, gre,
                         gid, rank, size)

            greater_id = gre[self.split.dim] > self.split.pos
            if greater_id:
                self.right.add_grid(gle, gre,
                         gid, rank, size)
        return



    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef insert_grid(self,
                    np.float64_t[:] gle,
                    np.float64_t[:] gre,
                    int grid_id,
                    int rank,
                    int size):
        if not should_i_build(self, rank, size):
            return

        # If we should continue to split based on parallelism, do so!
        if self.should_i_split(rank, size):
            self.geo_split(gle, gre, grid_id, rank, size)
            return

        cdef int contained = 1
        for i in range(3):
            if gle[i] > self.left_edge[i] or\
               gre[i] < self.right_edge[i]:
                contained *= 0

        if contained == 1:
            self.grid = grid_id
            assert(self.grid != -1)
            return

        # Split the grid
        cdef int check = self.split_grid(gle, gre, grid_id, rank, size)
        # If check is -1, then we have found a place where there are no choices.
        # Exit out and set the node to None.
        if check == -1:
            self.grid = -1
        return

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cpdef add_grids(self,
                        int ngrids,
                        np.float64_t[:,:] gles,
                        np.float64_t[:,:] gres,
                        np.int64_t[:] gids,
                        int rank,
                        int size):
        cdef int i, j, nless, ngreater, index
        cdef np.float64_t[:,:] less_gles, less_gres, greater_gles, greater_gres
        cdef np.int64_t[:] l_ids, g_ids
        if not should_i_build(self, rank, size):
            return

        if self._kd_is_leaf() == 1:
            self.insert_grids(ngrids, gles, gres, gids, rank, size)
            return

        less_ids = cvarray(format="q", shape=(ngrids,), itemsize=sizeof(np.int64_t))
        greater_ids = cvarray(format="q", shape=(ngrids,), itemsize=sizeof(np.int64_t))

        nless = 0
        ngreater = 0
        for i in range(ngrids):
            if gles[i, self.split.dim] < self.split.pos:
                less_ids[nless] = i
                nless += 1

            if gres[i, self.split.dim] > self.split.pos:
                greater_ids[ngreater] = i
                ngreater += 1

        #print('nless: %i' % nless)
        #print('ngreater: %i' % ngreater)

        if nless > 0:
            less_gles = cvarray(format="d", shape=(nless,3), itemsize=sizeof(np.float64_t))
            less_gres = cvarray(format="d", shape=(nless,3), itemsize=sizeof(np.float64_t))
            l_ids = cvarray(format="q", shape=(nless,), itemsize=sizeof(np.int64_t))

            for i in range(nless):
                index = less_ids[i]
                l_ids[i] = gids[index]
                for j in range(3):
                    less_gles[i,j] = gles[index,j]
                    less_gres[i,j] = gres[index,j]

            self.left.add_grids(nless, less_gles, less_gres,
                      l_ids, rank, size)

        if ngreater > 0:
            greater_gles = cvarray(format="d", shape=(ngreater,3), itemsize=sizeof(np.float64_t))
            greater_gres = cvarray(format="d", shape=(ngreater,3), itemsize=sizeof(np.float64_t))
            g_ids = cvarray(format="q", shape=(ngreater,), itemsize=sizeof(np.int64_t))

            for i in range(ngreater):
                index = greater_ids[i]
                g_ids[i] = gids[index]
                for j in range(3):
                    greater_gles[i,j] = gles[index,j]
                    greater_gres[i,j] = gres[index,j]

            self.right.add_grids(ngreater, greater_gles, greater_gres,
                      g_ids, rank, size)

        return

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int should_i_split(self, int rank, int size):
        if self.node_id < size and self.node_id > 0:
            return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void insert_grids(self,
                           int ngrids,
                           np.float64_t[:,:] gles,
                           np.float64_t[:,:] gres,
                           np.int64_t[:] gids,
                           int rank,
                           int size):

        if not should_i_build(self, rank, size) or ngrids == 0:
            return
        cdef int contained = 1
        cdef int check

        if ngrids == 1:
            # If we should continue to split based on parallelism, do so!
            if self.should_i_split(rank, size):
                self.geo_split(gles[0,:], gres[0,:], gids[0], rank, size)
                return

            for i in range(3):
                contained *= gles[0,i] <= self.left_edge[i]
                contained *= gres[0,i] >= self.right_edge[i]

            if contained == 1:
                # print('Node fully contained, setting to grid: %i' % gids[0])
                self.grid = gids[0]
                assert(self.grid != -1)
                return

        # Split the grids
        check = self.split_grids(ngrids, gles, gres, gids, rank, size)
        # If check is -1, then we have found a place where there are no choices.
        # Exit out and set the node to None.
        if check == -1:
            self.grid = -1
        return

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef split_grid(self,
                   np.float64_t[:] gle,
                   np.float64_t[:] gre,
                   int gid,
                   int rank,
                   int size):

        cdef int j
        cdef np.uint8_t[:] less_ids, greater_ids
        data = cvarray(format="d", shape=(1,2,3), itemsize=sizeof(np.float64_t))
        for j in range(3):
            data[0,0,j] = gle[j]
            data[0,1,j] = gre[j]

        less_ids = cvarray(format="B", shape=(1,), itemsize=sizeof(np.uint8_t))
        greater_ids = cvarray(format="B", shape=(1,), itemsize=sizeof(np.uint8_t))

        best_dim, split_pos, nless, ngreater = \
            kdtree_get_choices(1, data, self.left_edge, self.right_edge,
                              less_ids, greater_ids)

        # If best_dim is -1, then we have found a place where there are no choices.
        # Exit out and set the node to None.
        if best_dim == -1:
            return -1


        split = <Split *> malloc(sizeof(Split))
        split.dim = best_dim
        split.pos = split_pos

        # Create a Split
        self.divide(split)

        # Populate Left Node
        #print('Inserting left node', self.left_edge, self.right_edge)
        if nless == 1:
            self.left.insert_grid(gle, gre,
                         gid, rank, size)

        # Populate Right Node
        #print('Inserting right node', self.left_edge, self.right_edge)
        if ngreater == 1:
            self.right.insert_grid(gle, gre,
                         gid, rank, size)

        return 0

    #@cython.boundscheck(False)
    #@cython.wraparound(False)
    #@cython.cdivision(True)
    cdef int split_grids(self,
                           int ngrids,
                           np.float64_t[:,:] gles,
                           np.float64_t[:,:] gres,
                           np.int64_t[:] gids,
                           int rank,
                           int size):
        # Find a Split
        cdef int i, j, index
        cdef np.float64_t[:,:] less_gles, less_gres, greater_gles, greater_gres
        cdef np.int64_t[:] l_ids, g_ids
        if ngrids == 0: return 0

        data = cvarray(format="d", shape=(ngrids,2,3), itemsize=sizeof(np.float64_t))

        for i in range(ngrids):
            for j in range(3):
                data[i,0,j] = gles[i,j]
                data[i,1,j] = gres[i,j]

        less_ids = cvarray(format="B", shape=(ngrids,), itemsize=sizeof(np.uint8_t))
        greater_ids = cvarray(format="B", shape=(ngrids,), itemsize=sizeof(np.uint8_t))

        best_dim, split_pos, nless, ngreater = \
            kdtree_get_choices(ngrids, data, self.left_edge, self.right_edge,
                              less_ids, greater_ids)


        # If best_dim is -1, then we have found a place where there are no choices.
        # Exit out and set the node to None.
        if best_dim == -1:
            print('Failed to split grids.')
            return -1

        split = <Split *> malloc(sizeof(Split))
        split.dim = best_dim
        split.pos = split_pos

        # Create a Split
        self.divide(split)

        less_index = cvarray(format="q", shape=(ngrids,), itemsize=sizeof(np.int64_t))
        greater_index = cvarray(format="q", shape=(ngrids,), itemsize=sizeof(np.int64_t))

        nless = 0
        ngreater = 0
        for i in range(ngrids):
            if less_ids[i] == 1:
                less_index[nless] = i
                nless += 1

            if greater_ids[i] == 1:
                greater_index[ngreater] = i
                ngreater += 1

        if nless > 0:
            less_gles = cvarray(format="d", shape=(nless,3), itemsize=sizeof(np.float64_t))
            less_gres = cvarray(format="d", shape=(nless,3), itemsize=sizeof(np.float64_t))
            l_ids = cvarray(format="q", shape=(nless,), itemsize=sizeof(np.int64_t))

            for i in range(nless):
                index = less_index[i]
                l_ids[i] = gids[index]
                for j in range(3):
                    less_gles[i,j] = gles[index,j]
                    less_gres[i,j] = gres[index,j]

            # Populate Left Node
            #print('Inserting left node', self.left_edge, self.right_edge)
            self.left.insert_grids(nless, less_gles, less_gres,
                         l_ids, rank, size)

        if ngreater > 0:
            greater_gles = cvarray(format="d", shape=(ngreater,3), itemsize=sizeof(np.float64_t))
            greater_gres = cvarray(format="d", shape=(ngreater,3), itemsize=sizeof(np.float64_t))
            g_ids = cvarray(format="q", shape=(ngreater,), itemsize=sizeof(np.int64_t))

            for i in range(ngreater):
                index = greater_index[i]
                g_ids[i] = gids[index]
                for j in range(3):
                    greater_gles[i,j] = gles[index,j]
                    greater_gres[i,j] = gres[index,j]

            # Populate Right Node
            #print('Inserting right node', self.left_edge, self.right_edge)
            self.right.insert_grids(ngreater, greater_gles, greater_gres,
                         g_ids, rank, size)

        return 0

    cdef geo_split(self,
                   np.float64_t[:] gle,
                   np.float64_t[:] gre,
                   int grid_id,
                   int rank,
                   int size):
        cdef int big_dim = 0
        cdef int i
        cdef np.float64_t v, my_max = 0.0

        for i in range(3):
            v = gre[i] - gle[i]
            if v > my_max:
                my_max = v
                big_dim = i

        new_pos = (gre[big_dim] + gle[big_dim])/2.

        lnew_gle = cvarray(format="d", shape=(3,), itemsize=sizeof(np.float64_t))
        lnew_gre = cvarray(format="d", shape=(3,), itemsize=sizeof(np.float64_t))
        rnew_gle = cvarray(format="d", shape=(3,), itemsize=sizeof(np.float64_t))
        rnew_gre = cvarray(format="d", shape=(3,), itemsize=sizeof(np.float64_t))

        for j in range(3):
            lnew_gle[j] = gle[j]
            lnew_gre[j] = gre[j]
            rnew_gle[j] = gle[j]
            rnew_gre[j] = gre[j]

        split = <Split *> malloc(sizeof(Split))
        split.dim = big_dim
        split.pos = new_pos

        # Create a Split
        self.divide(split)

        #lnew_gre[big_dim] = new_pos
        # Populate Left Node
        #print('Inserting left node', self.left_edge, self.right_edge)
        self.left.insert_grid(lnew_gle, lnew_gre,
                grid_id, rank, size)

        #rnew_gle[big_dim] = new_pos
        # Populate Right Node
        #print('Inserting right node', self.left_edge, self.right_edge)
        self.right.insert_grid(rnew_gle, rnew_gre,
                grid_id, rank, size)
        return

    cdef void divide(self, Split * split):
        # Create a Split
        self.split = split

        cdef np.float64_t[:] le = np.empty(3, dtype='float64')
        cdef np.float64_t[:] re = np.empty(3, dtype='float64')

        cdef int i
        for i in range(3):
            le[i] = self.left_edge[i]
            re[i] = self.right_edge[i]
        re[split.dim] = split.pos

        self.left = Node(self, None, None,
                         le, re, self.grid,
                         _lchild_id(self.node_id))

        re[split.dim] = self.right_edge[split.dim]
        le[split.dim] = split.pos
        self.right = Node(self, None, None,
                          le, re, self.grid,
                          _rchild_id(self.node_id))

        return
    #
    def kd_sum_volume(self):
        cdef np.float64_t vol = 1.0
        if (self.left is None) and (self.right is None):
            if self.grid == -1:
                return 0.0
            for i in range(3):
                vol *= self.right_edge[i] - self.left_edge[i]
            return vol
        else:
            return self.left.kd_sum_volume() + self.right.kd_sum_volume()

    def kd_node_check(self):
        assert (self.left is None) == (self.right is None)
        if (self.left is None) and (self.right is None):
            if self.grid != -1:
                return np.prod(self.right_edge - self.left_edge)
            else: return 0.0
        else:
            return self.left.kd_node_check()+self.right.kd_node_check()

    def kd_is_leaf(self):
        cdef int has_l_child = self.left == None
        cdef int has_r_child = self.right == None
        assert has_l_child == has_r_child
        return has_l_child

    cdef int _kd_is_leaf(self):
        if self.left is None or self.right is None:
            return 1
        return 0

    def depth_traverse(self, max_node=None):
        '''
        Yields a depth-first traversal of the kd tree always going to
        the left child before the right.
        '''
        current = self
        previous = None
        if max_node is None:
            max_node = np.inf
        while current is not None:
            yield current
            current, previous = step_depth(current, previous)
            if current is None: break
            if current.node_id >= max_node:
                current = current.parent
                previous = current.right

    def depth_first_touch(self, max_node=None):
        '''
        Yields a depth-first traversal of the kd tree always going to
        the left child before the right.
        '''
        current = self
        previous = None
        if max_node is None:
            max_node = np.inf
        while current is not None:
            if previous is None or previous.parent != current:
                yield current
            current, previous = step_depth(current, previous)
            if current is None: break
            if current.node_id >= max_node:
                current = current.parent
                previous = current.right

    def breadth_traverse(self):
        '''
        Yields a breadth-first traversal of the kd tree always going to
        the left child before the right.
        '''
        current = self
        previous = None
        while current is not None:
            yield current
            current, previous = step_depth(current, previous)


    def viewpoint_traverse(self, viewpoint):
        '''
        Yields a viewpoint dependent traversal of the kd-tree.  Starts
        with nodes furthest away from viewpoint.
        '''

        current = self
        previous = None
        while current is not None:
            yield current
            current, previous = step_viewpoint(current, previous, viewpoint)

    cdef int point_in_node(self,
                           np.float64_t[:] point):
        cdef int i
        cdef int inside = 1
        for i in range(3):
            inside *= self.left_edge[i] <= point[i]
            inside *= self.right_edge[i] > point[i]
        return inside

    cdef Node _find_node(self, np.float64_t[:] point):
        while self._kd_is_leaf() == 0:
            if point[self.split.dim] < self.split.pos:
                self = self.left
            else:
                self = self.right
        return self

    def find_node(self,
                  np.float64_t[:] point):
        """
        Find the AMRKDTree node enclosing a position
        """
        assert(self.point_in_node(point))
        return self._find_node(point)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline np.int64_t _lchild_id(np.int64_t node_id):
    return (node_id<<1)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline np.int64_t _rchild_id(np.int64_t node_id):
    return (node_id<<1) + 1

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline np.int64_t _parent_id(np.int64_t node_id):
    return (node_id-1) >> 1

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef int should_i_build(Node node, int rank, int size):
    if (node.node_id < size) or (node.node_id >= 2*size):
        return 1
    elif node.node_id - size == rank:
        return 1
    else:
        return 0

def step_depth(Node current, Node previous):
    '''
    Takes a single step in the depth-first traversal
    '''
    if current._kd_is_leaf() == 1: # At a leaf, move back up
        previous = current
        current = current.parent

    elif current.parent is previous: # Moving down, go left first
        previous = current
        if current.left is not None:
            current = current.left
        elif current.right is not None:
            current = current.right
        else:
            current = current.parent

    elif current.left is previous: # Moving up from left, go right
        previous = current
        if current.right is not None:
            current = current.right
        else:
            current = current.parent

    elif current.right is previous: # Moving up from right child, move up
        previous = current
        current = current.parent

    return current, previous

def step_viewpoint(Node current,
                   Node previous,
                   viewpoint):
    '''
    Takes a single step in the viewpoint based traversal.  Always
    goes to the node furthest away from viewpoint first.
    '''
    if current._kd_is_leaf() == 1: # At a leaf, move back up
        previous = current
        current = current.parent
    elif current.split.dim is None: # This is a dead node
        previous = current
        current = current.parent

    elif current.parent is previous: # Moving down
        previous = current
        if viewpoint[current.split.dim] <= current.split.pos:
            if current.right is not None:
                current = current.right
            else:
                previous = current.right
        else:
            if current.left is not None:
                current = current.left
            else:
                previous = current.left

    elif current.right is previous: # Moving up from right
        previous = current
        if viewpoint[current.split.dim] <= current.split.pos:
            if current.left is not None:
                current = current.left
            else:
                current = current.parent
        else:
            current = current.parent

    elif current.left is previous: # Moving up from left child
        previous = current
        if viewpoint[current.split.dim] > current.split.pos:
            if current.right is not None:
                current = current.right
            else:
                current = current.parent
        else:
            current = current.parent

    return current, previous

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef kdtree_get_choices(int n_grids,
                        np.float64_t[:,:,:] data,
                        np.float64_t[:] l_corner,
                        np.float64_t[:] r_corner,
                        np.uint8_t[:] less_ids,
                        np.uint8_t[:] greater_ids,
                       ):
    cdef int i, j, k, dim, n_unique, best_dim, my_split
    cdef np.float64_t split
    cdef np.float64_t[:,:] uniquedims
    cdef np.float64_t[:] uniques
    uniquedims = cvarray(format="d", shape=(3, 2*n_grids), itemsize=sizeof(np.float64_t))
    my_max = 0
    my_split = 0
    best_dim = -1
    for dim in range(3):
        n_unique = 0
        uniques = uniquedims[dim]
        for i in range(n_grids):
            # Check for disqualification
            for j in range(2):
                # print("Checking against", i,j,dim,data[i,j,dim])
                if not (l_corner[dim] < data[i][j][dim] and
                        data[i][j][dim] < r_corner[dim]):
                    # print("Skipping ", data[i,j,dim], l_corner[dim], r_corner[dim])
                    continue
                skipit = 0
                # Add our left ...
                for k in range(n_unique):
                    if uniques[k] == data[i][j][dim]:
                        skipit = 1
                        # print("Identified", uniques[k], data[i,j,dim], n_unique)
                        break
                if skipit == 0:
                    uniques[n_unique] = data[i][j][dim]
                    n_unique += 1
        if n_unique > my_max:
            best_dim = dim
            my_max = n_unique
            my_split = (n_unique-1)/2
    if best_dim == -1:
        return -1, 0, 0, 0
    # I recognize how lame this is.
    cdef np.ndarray[np.float64_t, ndim=1] tarr = np.empty(my_max, dtype='float64')
    for i in range(my_max):
        # print("Setting tarr: ", i, uniquedims[best_dim][i])
        tarr[i] = uniquedims[best_dim][i]
    tarr.sort()
    split = tarr[my_split]
    cdef int nless=0, ngreater=0
    for i in range(n_grids):
        if data[i][0][best_dim] < split:
            less_ids[i] = 1
            nless += 1
        else:
            less_ids[i] = 0
        if data[i][1][best_dim] > split:
            greater_ids[i] = 1
            ngreater += 1
        else:
            greater_ids[i] = 0

    # Return out unique values
    return best_dim, split, nless, ngreater



import os
import warnings
from functools import wraps
from typing import TYPE_CHECKING, TypeVar

import matplotlib as mpl
from matplotlib.ticker import SymmetricalLogLocator
from more_itertools import always_iterable

from yt.config import ytcfg

if TYPE_CHECKING:
    from matplotlib.backend_bases import FigureCanvasBase


_DEFAULT_FONT_PROPERTIES = None


def get_default_font_properties():
    global _DEFAULT_FONT_PROPERTIES
    if _DEFAULT_FONT_PROPERTIES is None:
        import importlib.resources as importlib_resources

        _yt_style = mpl.rc_params_from_file(
            importlib_resources.files("yt") / "default.mplstyle",
            use_default_template=False,
        )
        _DEFAULT_FONT_PROPERTIES = {
            "family": _yt_style["font.family"][0],
            "math_fontfamily": _yt_style["mathtext.fontset"],
        }

    return _DEFAULT_FONT_PROPERTIES


def _get_supported_image_file_formats():
    from matplotlib.backend_bases import FigureCanvasBase

    return frozenset(FigureCanvasBase.get_supported_filetypes().keys())


def _get_supported_canvas_classes():
    from matplotlib.backends.backend_agg import FigureCanvasAgg
    from matplotlib.backends.backend_pdf import FigureCanvasPdf
    from matplotlib.backends.backend_ps import FigureCanvasPS
    from matplotlib.backends.backend_svg import FigureCanvasSVG

    return frozenset(
        (FigureCanvasAgg, FigureCanvasPdf, FigureCanvasPS, FigureCanvasSVG)
    )


def get_canvas_class(suffix: str) -> type["FigureCanvasBase"]:
    s = suffix.removeprefix(".")
    if s not in _get_supported_image_file_formats():
        raise ValueError(f"Unsupported file format '{suffix}'.")
    for cls in _get_supported_canvas_classes():
        if s in cls.get_supported_filetypes():
            return cls
    raise RuntimeError(
        "Something went terribly wrong. "
        f"File extension '{suffix}' is supposed to be supported "
        "but no compatible backend was found."
    )


def validate_image_name(filename, suffix: str | None = None) -> str:
    """
    Build a valid image filename with a specified extension (default to png).
    The suffix parameter is ignored if the input filename has a valid extension already.
    Otherwise, suffix is appended to the filename, replacing any existing extension.
    """
    name, psuffix = os.path.splitext(filename)
    psuffix = psuffix.removeprefix(".")

    if suffix is not None:
        suffix = suffix.removeprefix(".")

    if psuffix in _get_supported_image_file_formats():
        if suffix in _get_supported_image_file_formats() and suffix != psuffix:
            warnings.warn(
                f"Received two valid image formats {psuffix!r} (from filename) "
                f"and {suffix!r} (from suffix). The former is ignored.",
                stacklevel=2,
            )
            return f"{name}.{suffix}"
        return str(filename)

    if suffix is None:
        suffix = "png"

    if suffix not in _get_supported_image_file_formats():
        raise ValueError(f"Unsupported file format {suffix!r}")

    return f"{filename}.{suffix}"


def get_canvas(figure, filename):
    name, suffix = os.path.splitext(filename)

    if not suffix:
        raise ValueError(
            f"Can not determine canvas class from filename '{filename}' "
            f"without an extension."
        )
    return get_canvas_class(suffix)(figure)


def invalidate_plot(f):
    @wraps(f)
    def newfunc(self, *args, **kwargs):
        retv = f(self, *args, **kwargs)
        self._plot_valid = False
        return retv

    return newfunc


def invalidate_data(f):
    @wraps(f)
    def newfunc(self, *args, **kwargs):
        retv = f(self, *args, **kwargs)
        self._data_valid = False
        self._plot_valid = False
        return retv

    return newfunc


def invalidate_figure(f):
    @wraps(f)
    def newfunc(self, *args, **kwargs):
        retv = f(self, *args, **kwargs)
        for field in self.plots.keys():
            self.plots[field].figure = None
            self.plots[field].axes = None
            self.plots[field].cax = None
        self._setup_plots()
        return retv

    return newfunc


def validate_plot(f):
    @wraps(f)
    def newfunc(self, *args, **kwargs):
        # TODO: _profile_valid and _data_valid seem to play very similar roles,
        # there's probably room to abstract these into a common operation
        if hasattr(self, "_data_valid") and not self._data_valid:
            self._recreate_frb()
        if hasattr(self, "_profile_valid") and not self._profile_valid:
            self._recreate_profile()
        if not self._plot_valid:
            # it is the responsibility of _setup_plots to
            # call plot.run_callbacks()
            self._setup_plots()
        retv = f(self, *args, **kwargs)
        return retv

    return newfunc


T = TypeVar("T", tuple, list)


def _swap_axes_extents(extent: T) -> T:
    """
    swaps the x and y extent values, preserving type of extent

    Parameters
    ----------
    extent : sequence of four unyt quantities
        the current 4-element tuple or list of unyt quantities describing the
        plot extent. extent = (xmin, xmax, ymin, ymax).

    Returns
    -------
    tuple or list
        the extent axes swapped, now with (ymin, ymax, xmin, xmax).

    """
    extent_swapped = [extent[2], extent[3], extent[0], extent[1]]
    return type(extent)(extent_swapped)


def _swap_arg_pair_order(*args):
    """
    flips adjacent argument pairs, useful for swapping x-y plot arguments

    Parameters
    ----------
    *args
        argument pairs, must have an even number of *args

    Returns
    -------
    tuple
        args  with order of pairs switched, i.e,:

        _swap_arg_pair_order(x, y, px, py) returns:
            y, x, py, px

    """

    if len(args) % 2 != 0:
        raise TypeError("Number of arguments must be even.")
    n_pairs = len(args) // 2
    new_args = []
    for i in range(n_pairs):
        x_id = i * 2
        new_args.append(args[x_id + 1])
        new_args.append(args[x_id])
    return tuple(new_args)


class _MPL38_SymmetricalLogLocator(SymmetricalLogLocator):
    # Backporting behaviour from matplotlib 3.8 (in development at the time of writing)
    # see https://github.com/matplotlib/matplotlib/pull/25970

    def __init__(self, *args, **kwargs):
        if mpl.__version_info__ >= (3, 8):
            raise RuntimeError(
                "_MPL38_SymmetricalLogLocator is not needed with matplotlib>=3.8"
            )
        super().__init__(*args, **kwargs)

    def tick_values(self, vmin, vmax):
        linthresh = self._linthresh
        if vmax < vmin:
            vmin, vmax = vmax, vmin
        if -linthresh <= vmin < vmax <= linthresh:
            # only the linear range is present
            return sorted({vmin, 0, vmax})

        return super().tick_values(vmin, vmax)


def get_default_from_config(data_source, *, field, keys, defaults):
    _keys = list(always_iterable(keys))
    _defaults = list(always_iterable(defaults))

    ftype, fname = data_source._determine_fields(field)[0]
    ret = [
        ytcfg.get_most_specific("plot", ftype, fname, key, fallback=default)
        for key, default in zip(_keys, _defaults, strict=True)
    ]
    if len(ret) == 1:
        return ret[0]
    else:
        return ret


def _get_units_label(units: str) -> str:
    if r"\frac" in units:
        return rf"$\ \ \left({units}\right)$"
    elif units:
        return rf"$\ \ ({units})$"
    else:
        return ""



import os

import numpy as np

from yt.frontends.gadget.data_structures import GadgetHDF5Dataset
from yt.utilities.cosmology import Cosmology
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import GizmoFieldInfo


class GizmoDataset(GadgetHDF5Dataset):
    _load_requirements = ["h5py"]
    _field_info_class = GizmoFieldInfo

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        need_groups = ["Header"]
        veto_groups = ["Config", "Constants", "FOF", "Group", "Subhalo"]
        valid = True
        valid_fname = filename
        # If passed arg is a directory, look for the .0 file in that dir
        if os.path.isdir(filename):
            valid_files = []
            for f in os.listdir(filename):
                fname = os.path.join(filename, f)
                fext = os.path.splitext(fname)[-1]
                if (
                    (".0" in f)
                    and (fext not in {".ewah", ".kdtree"})
                    and os.path.isfile(fname)
                ):
                    valid_files.append(fname)
            if len(valid_files) == 0:
                valid = False
            elif len(valid_files) > 1:
                valid = False
            else:
                valid_fname = valid_files[0]
        try:
            fh = h5py.File(valid_fname, mode="r")
            valid = all(ng in fh["/"] for ng in need_groups) and not any(
                vg in fh["/"] for vg in veto_groups
            )
            # From Apr 2021, 7f1f06f, public gizmo includes a header variable
            # GIZMO_version, which is set to the year of the most recent commit
            # We should prefer this to checking the metallicity, which might
            # not exist
            if "GIZMO_version" not in fh["/Header"].attrs:
                dmetal = "/PartType0/Metallicity"
                if dmetal not in fh or (
                    fh[dmetal].ndim > 1 and fh[dmetal].shape[1] < 11
                ):
                    valid = False
            fh.close()
        except Exception:
            valid = False
        return valid

    def _set_code_unit_attributes(self):
        super()._set_code_unit_attributes()

    def _parse_parameter_file(self):
        hvals = self._get_hvals()

        self.dimensionality = 3
        self.refine_by = 2
        self.parameters["HydroMethod"] = "sph"
        # Set standard values

        # We may have an overridden bounding box.
        if self.domain_left_edge is None and hvals["BoxSize"] != 0:
            self.domain_left_edge = np.zeros(3, "float64")
            self.domain_right_edge = np.ones(3, "float64") * hvals["BoxSize"]

        self.domain_dimensions = np.ones(3, "int32")
        self._periodicity = (True, True, True)

        self.cosmological_simulation = 1

        self.current_redshift = hvals.get("Redshift", 0.0)
        if "Redshift" not in hvals:
            mylog.info("Redshift is not set in Header. Assuming z=0.")

        if "ComovingIntegrationOn" in hvals:
            # In 1d8479, Nov 2020, public GIZMO updated the names of the Omegas
            # to include an _, added baryons and radiation and added the
            # ComovingIntegrationOn field. ComovingIntegrationOn is always set,
            # but the Omega's are only included if ComovingIntegrationOn is true
            mylog.debug("Reading cosmological parameters using post-1d8479 format")
            self.cosmological_simulation = hvals["ComovingIntegrationOn"]
            if self.cosmological_simulation:
                self.omega_lambda = hvals["Omega_Lambda"]
                self.omega_matter = hvals["Omega_Matter"]
                self.omega_baryon = hvals["Omega_Baryon"]
                self.omega_radiation = hvals["Omega_Radiation"]
            self.hubble_constant = hvals["HubbleParam"]
        elif "OmegaLambda" in hvals:
            # Should still support GIZMO versions prior to 1d8479 too
            mylog.info(
                "ComovingIntegrationOn does not exist, falling back to OmegaLambda",
            )
            self.omega_lambda = hvals["OmegaLambda"]
            self.omega_matter = hvals["Omega0"]
            self.hubble_constant = hvals["HubbleParam"]
            self.cosmological_simulation = self.omega_lambda != 0.0
        else:
            # If these are not set it is definitely not a cosmological dataset.
            mylog.debug("No cosmological information found, assuming defaults")
            self.omega_lambda = 0.0
            self.omega_matter = 0.0  # Just in case somebody asks for it.
            self.cosmological_simulation = 0
            # Hubble is set below for Omega Lambda = 0.

        if not self.cosmological_simulation:
            mylog.info(
                "ComovingIntegrationOn != 1 or (not found "
                "and OmegaLambda is 0.0), so we are turning off Cosmology.",
            )
            self.hubble_constant = 1.0  # So that scaling comes out correct
            self.current_redshift = 0.0
            # This may not be correct.
            self.current_time = hvals["Time"]
        else:
            # Now we calculate our time based on the cosmology, because in
            # ComovingIntegration hvals["Time"] will in fact be the expansion
            # factor, not the actual integration time, so we re-calculate
            # global time from our Cosmology.
            cosmo = Cosmology(
                hubble_constant=self.hubble_constant,
                omega_matter=self.omega_matter,
                omega_lambda=self.omega_lambda,
            )
            self.current_time = cosmo.lookback_time(self.current_redshift, 1e6)
            mylog.info(
                "Calculating time from %0.3e to be %0.3e seconds",
                hvals["Time"],
                self.current_time,
            )
        self.parameters = hvals

        prefix = os.path.join(self.directory, self.basename.split(".", 1)[0])

        if hvals["NumFiles"] > 1:
            self.filename_template = f"{prefix}.%(num)s{self._suffix}"
        else:
            self.filename_template = self.parameter_filename

        self.file_count = hvals["NumFiles"]



"""
Definitions for the lens code




"""


import numpy as np

cimport cython
cimport numpy as np
from libc.math cimport (
    M_PI,
    acos,
    asin,
    atan,
    atan2,
    cos,
    exp,
    fabs,
    floor,
    log2,
    sin,
    sqrt,
)

from yt.utilities.lib.fp_utils cimport fclip, fmax, fmin, i64clip, iclip, imax, imin

from .image_samplers cimport (
    ImageSampler,
    calculate_extent_function,
    generate_vector_info_function,
)
from .vec3_ops cimport L2_norm, dot, fma, subtract
from .volume_container cimport VolumeContainer


cdef extern from "platform_dep.h":
    long int lrint(double x) noexcept nogil

cdef extern from "limits.h":
    cdef int SHRT_MAX

cdef generate_vector_info_function generate_vector_info_plane_parallel
cdef generate_vector_info_function generate_vector_info_null
cdef calculate_extent_function calculate_extent_plane_parallel
cdef calculate_extent_function calculate_extent_perspective
cdef calculate_extent_function calculate_extent_null



"""
A cython implementation of the bounded priority queue

This is a priority queue that only keeps track of smallest k values that have
been added to it.


"""

import numpy as np

cimport numpy as np


cdef class BoundedPriorityQueue:
    cdef public np.float64_t[:] heap
    cdef np.float64_t* heap_ptr
    cdef public np.int64_t[:] pids
    cdef np.int64_t* pids_ptr
    cdef int use_pids

    cdef np.intp_t size
    cdef np.intp_t max_elements

    cdef int max_heapify(self, np.intp_t index) except -1 nogil
    cdef int propagate_up(self, np.intp_t index) except -1 nogil
    cdef int add(self, np.float64_t val) except -1 nogil
    cdef int add_pid(self, np.float64_t val, np.int64_t pid) except -1 nogil
    cdef int heap_append(self, np.float64_t val, np.int64_t ind) except -1 nogil
    cdef np.float64_t extract_max(self) except -1 nogil
    cdef int validate_heap(self) except -1 nogil

cdef class NeighborList:
    cdef public np.float64_t[:] data
    cdef np.float64_t* data_ptr
    cdef public np.int64_t[:] pids
    cdef np.int64_t* pids_ptr
    cdef np.intp_t size
    cdef np.intp_t _max_size

    cdef int _update_memview(self) except -1
    cdef int _extend(self) except -1 nogil
    cdef int add_pid(self, np.float64_t val, np.int64_t ind) except -1 nogil



import numpy as np

import yt

# load data
fn = "IsolatedGalaxy/galaxy0030/galaxy0030"
ds = yt.load(fn)

# This is the number of frames to make -- below, you can see how this is used.
n_frames = 5

# This is the minimum size in smallest_dx of our last frame.
# Usually it should be set to something like 400, but for THIS
# dataset, we actually don't have that great of resolution.
min_dx = 40

frame_template = "frame_%05i"  # Template for frame filenames

p = yt.SlicePlot(ds, "z", ("gas", "density"))  # Add our slice, along z
p.annotate_contour(("gas", "temperature"))  # We'll contour in temperature

# What we do now is a bit fun.  "enumerate" returns a tuple for every item --
# the index of the item, and the item itself.  This saves us having to write
# something like "i = 0" and then inside the loop "i += 1" for ever loop.  The
# argument to enumerate is the 'logspace' function, which takes a minimum and a
# maximum and the number of items to generate.  It returns 10^power of each
# item it generates.

for i, v in enumerate(
    np.logspace(0, np.log10(ds.index.get_smallest_dx() * min_dx), n_frames)
):
    # We set our width as necessary for this frame
    p.set_width(v, "unitary")
    # save
    p.save(frame_template % (i))



import numpy as np

import yt
from yt.units import kpc
from yt.visualization.volume_rendering.api import LineSource

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

sc = yt.create_scene(ds)

np.random.seed(1234567)

nlines = 50
vertices = (np.random.random([nlines, 2, 3]) - 0.5) * 200 * kpc
colors = np.random.random([nlines, 4])
colors[:, 3] = 0.1

lines = LineSource(vertices, colors)
sc.add_source(lines)

sc.camera.width = 300 * kpc

sc.save(sigma_clip=4.0)



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.fields.tensor_fields import setup_stress_energy_ideal
from yt.funcs import mylog

from .cfields import SRHDFields

b_units = "code_magnetic"
pre_units = "code_mass / (code_length*code_time**2)"
erg_units = "code_mass / (code_length*code_time**2)"
rho_units = "code_mass / code_length**3"
mom_units = "code_mass / (code_length**2*code_time)"
vel_units = "code_velocity"
pot_units = "code_length**2/code_time**2"

psi_units = "code_mass**(1/2) / code_length**(3/2)"


class GAMERFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        # hydro fields on disk (GAMER outputs conservative variables)
        ("Dens", (rho_units, [], None)),
        ("MomX", (mom_units, ["momentum_density_x"], None)),
        ("MomY", (mom_units, ["momentum_density_y"], None)),
        ("MomZ", (mom_units, ["momentum_density_z"], None)),
        ("Engy", (erg_units, [], None)),
        ("CRay", (erg_units, ["cosmic_ray_energy_density"], None)),
        ("Pote", (pot_units, ["gravitational_potential"], None)),
        ("Pres", (pre_units, ["pressure"], None)),
        ("Temp", ("code_temperature", ["temperature"], None)),
        ("Enth", (pot_units, ["specific_reduced_enthalpy"], None)),
        ("Mach", ("dimensionless", ["mach_number"], None)),
        ("Cs", (vel_units, ["sound_speed"], None)),
        ("DivVel", ("1/code_time", ["velocity_divergence"], None)),
        # MHD fields on disk (CC=cell-centered)
        ("CCMagX", (b_units, [], "B_x")),
        ("CCMagY", (b_units, [], "B_y")),
        ("CCMagZ", (b_units, [], "B_z")),
        # psiDM fields on disk
        ("Real", (psi_units, ["psidm_real_part"], None)),
        ("Imag", (psi_units, ["psidm_imaginary_part"], None)),
        # particle fields on disk (deposited onto grids)
        ("ParDens", (rho_units, ["particle_density_on_grid"], None)),
        ("TotalDens", (rho_units, ["total_density_on_grid"], None)),
    )

    known_particle_fields: KnownFieldsT = (
        ("ParMass", ("code_mass", ["particle_mass"], None)),
        ("ParPosX", ("code_length", ["particle_position_x"], None)),
        ("ParPosY", ("code_length", ["particle_position_y"], None)),
        ("ParPosZ", ("code_length", ["particle_position_z"], None)),
        ("ParVelX", ("code_velocity", ["particle_velocity_x"], None)),
        ("ParVelY", ("code_velocity", ["particle_velocity_y"], None)),
        ("ParVelZ", ("code_velocity", ["particle_velocity_z"], None)),
        ("ParCreTime", ("code_time", ["particle_creation_time"], None)),
    )

    def __init__(self, ds, field_list):
        super().__init__(ds, field_list)

    # add primitive and other derived variables
    def setup_fluid_fields(self):
        pc = self.ds.units.physical_constants
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        unit_system = self.ds.unit_system
        unit_system.registry = self.ds.unit_registry  # TODO: Why do I need this?!

        if self.ds.opt_unit:
            temp_conv = pc.kb / (self.ds.mu * pc.mh)
        else:
            temp_conv = (
                self.ds.arr(1.0, "code_velocity**2/code_temperature") / self.ds.mu
            )

        if self.ds.srhd:
            if self.ds.opt_unit:
                c2 = pc.clight * pc.clight
            else:
                c2 = self.ds.arr(1.0, "code_velocity**2")
            invc2 = 1.0 / c2

            if ("gamer", "Temp") not in self.field_list:
                mylog.warning(
                    'The temperature field "Temp" is not present in the dataset. Most '
                    'SRHD fields will not be available!! Please set "OPT__OUTPUT_TEMP '
                    '= 1" in Input__Parameter and re-run the simulation.'
                )
            if ("gamer", "Enth") not in self.field_list:
                mylog.warning(
                    'The reduced enthalpy field "Enth" is not present in the dataset. '
                    "Most SRHD fields will not be available!! Please set "
                    '"OPT__OUTPUT_ENTHALPY = 1" in Input__Parameter and re-run the '
                    "simulation."
                )

            # EOS functions
            gamma = self.ds.gamma if self.ds.eos == 1 else 0.0
            fgen = SRHDFields(self.ds.eos, gamma)

            # temperature fraction (kT/mc^2)
            def _temp_fraction(field, data):
                return data["gamer", "Temp"] * temp_conv * invc2

            self.add_field(
                ("gas", "temp_fraction"),
                function=_temp_fraction,
                sampling_type="cell",
                units="",
            )

            # specific enthalpy
            def _specific_enthalpy(field, data):
                return data["gas", "specific_reduced_enthalpy"] + c2

            self.add_field(
                ("gas", "specific_enthalpy"),
                function=_specific_enthalpy,
                sampling_type="cell",
                units=unit_system["specific_energy"],
            )

            # sound speed
            if ("gamer", "Cs") not in self.field_list:

                def _sound_speed(field, data):
                    out = fgen.sound_speed(
                        data["gas", "temp_fraction"].d,
                        data["gamer", "Enth"].d,
                    )
                    return data.ds.arr(out, "code_velocity").to(unit_system["velocity"])

                self.add_field(
                    ("gas", "sound_speed"),
                    sampling_type="cell",
                    function=_sound_speed,
                    units=unit_system["velocity"],
                )

            # ratio of specific heats (gamma)
            def _gamma(field, data):
                out = fgen.gamma_field(data["gas", "temp_fraction"].d)
                return data.ds.arr(out, "dimensionless")

            self.add_field(
                ("gas", "gamma"), sampling_type="cell", function=_gamma, units=""
            )

            # reduced total energy density
            self.alias(
                ("gas", "reduced_total_energy_density"),
                ("gamer", "Engy"),
                units=unit_system["pressure"],
            )

            # total energy density
            def _total_energy_density(field, data):
                return data["gamer", "Engy"] + data["gamer", "Dens"] * c2

            self.add_field(
                ("gas", "total_energy_density"),
                sampling_type="cell",
                function=_total_energy_density,
                units=unit_system["pressure"],
            )

            # coordinate frame density
            self.alias(
                ("gas", "frame_density"),
                ("gamer", "Dens"),
                units=unit_system["density"],
            )

            # 4-velocity spatial components
            def four_velocity_xyz(u):
                def _four_velocity(field, data):
                    out = fgen.four_velocity_xyz(
                        data["gamer", "Dens"].d,
                        data["gamer", f"Mom{u.upper()}"].d,
                        data["gamer", "Enth"].d,
                    )
                    return data.ds.arr(out, "code_velocity").to(unit_system["velocity"])

                return _four_velocity

            for u in "xyz":
                self.add_field(
                    ("gas", f"four_velocity_{u}"),
                    sampling_type="cell",
                    function=four_velocity_xyz(u),
                    units=unit_system["velocity"],
                )

            # lorentz factor
            if ("gamer", "Lrtz") in self.field_list:

                def _lorentz_factor(field, data):
                    return data["gamer", "Lrtz"]

            else:

                def _lorentz_factor(field, data):
                    out = fgen.lorentz_factor(
                        data["gamer", "Dens"].d,
                        data["gamer", "MomX"].d,
                        data["gamer", "MomY"].d,
                        data["gamer", "MomZ"].d,
                        data["gamer", "Enth"].d,
                    )
                    return data.ds.arr(out, "dimensionless")

            self.add_field(
                ("gas", "lorentz_factor"),
                sampling_type="cell",
                function=_lorentz_factor,
                units="",
            )

            # density
            def _density(field, data):
                return data["gamer", "Dens"] / data["gas", "lorentz_factor"]

            self.add_field(
                ("gas", "density"),
                sampling_type="cell",
                function=_density,
                units=unit_system["density"],
            )

            # pressure
            def _pressure(field, data):
                p = data["gas", "density"] * data["gas", "temp_fraction"]
                return p * c2

            # thermal energy per mass (i.e., specific)
            def _specific_thermal_energy(field, data):
                return (
                    data["gas", "specific_reduced_enthalpy"]
                    - c2 * data["gas", "temp_fraction"]
                )

            # total energy per mass
            def _specific_total_energy(field, data):
                return data["gas", "total_energy_density"] / data["gas", "density"]

            # kinetic energy density
            def _kinetic_energy_density(field, data):
                out = fgen.kinetic_energy_density(
                    data["gamer", "Dens"].d,
                    data["gamer", "MomX"].d,
                    data["gamer", "MomY"].d,
                    data["gamer", "MomZ"].d,
                    data["gas", "temp_fraction"].d,
                    data["gamer", "Enth"].d,
                )
                return data.ds.arr(out, erg_units).to(unit_system["pressure"])

            self.add_field(
                ("gas", "kinetic_energy_density"),
                sampling_type="cell",
                function=_kinetic_energy_density,
                units=unit_system["pressure"],
            )

            # Mach number
            if ("gamer", "Mach") not in self.field_list:

                def _mach_number(field, data):
                    out = fgen.mach_number(
                        data["gamer", "Dens"].d,
                        data["gamer", "MomX"].d,
                        data["gamer", "MomY"].d,
                        data["gamer", "MomZ"].d,
                        data["gas", "temp_fraction"].d,
                        data["gamer", "Enth"].d,
                    )
                    return data.ds.arr(out, "dimensionless")

                self.add_field(
                    ("gas", "mach_number"),
                    sampling_type="cell",
                    function=_mach_number,
                    units="",
                )

            setup_stress_energy_ideal(self)

        else:  # not RHD
            # density
            self.alias(
                ("gas", "density"), ("gamer", "Dens"), units=unit_system["density"]
            )

            self.alias(
                ("gas", "total_energy_density"),
                ("gamer", "Engy"),
                units=unit_system["pressure"],
            )

            # ====================================================
            # note that yt internal fields assume
            #    [specific_thermal_energy]   = [energy per mass]
            #    [kinetic_energy_density]    = [energy per volume]
            #    [magnetic_energy_density]   = [energy per volume]
            # and we further adopt
            #    [specific_total_energy]     = [energy per mass]
            #    [total_energy_density]      = [energy per volume]
            # ====================================================

            # thermal energy per volume
            def et(data):
                ek = (
                    0.5
                    * (
                        data["gamer", "MomX"] ** 2
                        + data["gamer", "MomY"] ** 2
                        + data["gamer", "MomZ"] ** 2
                    )
                    / data["gamer", "Dens"]
                )
                Et = data["gamer", "Engy"] - ek
                if self.ds.mhd:
                    # magnetic_energy is a yt internal field
                    Et -= data["gas", "magnetic_energy_density"]
                if getattr(self.ds, "gamma_cr", None):
                    # cosmic rays are included in this dataset
                    Et -= data["gas", "cosmic_ray_energy_density"]
                return Et

            # thermal energy per mass (i.e., specific)
            def _specific_thermal_energy(field, data):
                return et(data) / data["gamer", "Dens"]

            # total energy per mass
            def _specific_total_energy(field, data):
                return data["gamer", "Engy"] / data["gamer", "Dens"]

            # pressure
            def _pressure(field, data):
                return et(data) * (data.ds.gamma - 1.0)

        # velocity
        def velocity_xyz(v):
            if ("gamer", f"Vel{v.upper()}") in self.field_list:

                def _velocity(field, data):
                    return data.ds.arr(
                        data["gamer", f"Vel{v.upper()}"].d, "code_velocity"
                    ).to(unit_system["velocity"])

            elif self.ds.srhd:

                def _velocity(field, data):
                    return (
                        data["gas", f"four_velocity_{v}"]
                        / data["gas", "lorentz_factor"]
                    )

            else:

                def _velocity(field, data):
                    return data["gas", f"momentum_density_{v}"] / data["gas", "density"]

            return _velocity

        for v in "xyz":
            self.add_field(
                ("gas", f"velocity_{v}"),
                sampling_type="cell",
                function=velocity_xyz(v),
                units=unit_system["velocity"],
            )

        if ("gamer", "Pres") not in self.field_list:
            self.add_field(
                ("gas", "pressure"),
                sampling_type="cell",
                function=_pressure,
                units=unit_system["pressure"],
            )

        self.add_field(
            ("gas", "specific_thermal_energy"),
            sampling_type="cell",
            function=_specific_thermal_energy,
            units=unit_system["specific_energy"],
        )

        def _thermal_energy_density(field, data):
            return data["gas", "density"] * data["gas", "specific_thermal_energy"]

        self.add_field(
            ("gas", "thermal_energy_density"),
            sampling_type="cell",
            function=_thermal_energy_density,
            units=unit_system["pressure"],
        )

        self.add_field(
            ("gas", "specific_total_energy"),
            sampling_type="cell",
            function=_specific_total_energy,
            units=unit_system["specific_energy"],
        )

        if getattr(self.ds, "gamma_cr", None):

            def _cr_pressure(field, data):
                return (data.ds.gamma_cr - 1.0) * data[
                    "gas", "cosmic_ray_energy_density"
                ]

            self.add_field(
                ("gas", "cosmic_ray_pressure"),
                _cr_pressure,
                sampling_type="cell",
                units=self.ds.unit_system["pressure"],
            )

        # mean molecular weight
        if hasattr(self.ds, "mu"):

            def _mu(field, data):
                return data.ds.mu * data["index", "ones"]

            self.add_field(
                ("gas", "mean_molecular_weight"),
                sampling_type="cell",
                function=_mu,
                units="",
            )

        # temperature
        if ("gamer", "Temp") not in self.field_list:

            def _temperature(field, data):
                return data["gas", "pressure"] / (data["gas", "density"] * temp_conv)

            self.add_field(
                ("gas", "temperature"),
                sampling_type="cell",
                function=_temperature,
                units=unit_system["temperature"],
            )

        # magnetic field aliases --> magnetic_field_x/y/z
        if self.ds.mhd:
            setup_magnetic_field_aliases(self, "gamer", [f"CCMag{v}" for v in "XYZ"])

    def setup_particle_fields(self, ptype):
        super().setup_particle_fields(ptype)



import numpy as np
import pytest
import unyt
from numpy.testing import assert_allclose

from yt.testing import (
    assert_rel_equal,
    cubicspline_python,
    fake_amr_ds,
    fake_sph_flexible_grid_ds,
    fake_sph_grid_ds,
    integrate_kernel,
    requires_module_pytest,
)
from yt.visualization.api import OffAxisProjectionPlot, ProjectionPlot


@pytest.mark.parametrize("weighted", [True, False])
@pytest.mark.parametrize("periodic", [True, False])
@pytest.mark.parametrize("depth", [None, (1.0, "cm"), (0.5, "cm")])
@pytest.mark.parametrize("shiftcenter", [False, True])
@pytest.mark.parametrize("northvector", [None, (1.0e-4, 1.0, 0.0)])
def test_sph_proj_general_offaxis(
    northvector: tuple[float, float, float] | None,
    shiftcenter: bool,
    depth: tuple[float, str] | None,
    periodic: bool,
    weighted: bool,
) -> None:
    """
    Same as the on-axis projections, but we rotate the basis vectors
    to test whether roations are handled ok. the rotation is chosen to
    be small so that in/exclusion of particles within bboxes, etc.
    works out the same way.
    We just send lines of sight through pixel centers for convenience.
    Particles at [0.5, 1.5, 2.5] (in each coordinate)
    smoothing lengths 0.25
    all particles have mass 1., density 1.5,
    except the single center particle, with mass 2., density 3.

    Parameters:
    -----------
    northvector: tuple
        y-axis direction in the final plot (direction vector)
    shiftcenter: bool
        shift the coordinates to center the projection on.
        (The grid is offset to this same center)
    depth: float or None
        depth of the projection slice
    periodic: bool
        assume periodic boundary conditions, or not
    weighted: bool
        make a weighted projection (density-weighted density), or not

    Returns:
    --------
    None
    """
    if shiftcenter:
        center = np.array((0.625, 0.625, 0.625))  # cm
    else:
        center = np.array((1.5, 1.5, 1.5))  # cm
    bbox = unyt.unyt_array(np.array([[0.0, 3.0], [0.0, 3.0], [0.0, 3.0]]), "cm")
    hsml_factor = 0.5
    unitrho = 1.5

    # test correct centering, particle selection
    def makemasses(i, j, k):
        if i == j == k == 1:
            return 2.0
        else:
            return 1.0

    # result shouldn't depend explicitly on the center if we re-center
    # the data, unless we get cut-offs in the non-periodic case
    # *almost* the z-axis
    # try to make sure dl differences from periodic wrapping are small
    epsilon = 1e-4
    projaxis = np.array([epsilon, 0.00, np.sqrt(1.0 - epsilon**2)])
    e1dir = projaxis / np.sqrt(np.sum(projaxis**2))
    # TODO: figure out other (default) axes for basis vectors here
    if northvector is None:
        e2dir = np.array([0.0, 1.0, 0.0])
    else:
        e2dir = np.asarray(northvector)
    e2dir = e2dir - np.sum(e1dir * e2dir) * e2dir  # orthonormalize
    e2dir /= np.sqrt(np.sum(e2dir**2))
    e3dir = np.cross(e1dir, e2dir)

    ds = fake_sph_flexible_grid_ds(
        hsml_factor=hsml_factor,
        nperside=3,
        periodic=periodic,
        offsets=np.full(3, 0.5),
        massgenerator=makemasses,
        unitrho=unitrho,
        bbox=bbox.v,
        recenter=center,
        e1hat=e1dir,
        e2hat=e2dir,
        e3hat=e3dir,
    )

    source = ds.all_data()
    # couple to dataset -> right unit registry
    center = ds.arr(center, "cm")
    # print('position:\n', source['gas','position'])

    # m / rho, factor 1. / hsml**2 is included in the kernel integral
    # (density is adjusted, so same for center particle)
    prefactor = 1.0 / unitrho  # / (0.5 * 0.5)**2
    dl_cen = integrate_kernel(cubicspline_python, 0.0, 0.25)

    if weighted:
        toweight_field = ("gas", "density")
    else:
        toweight_field = None
    # we don't actually want a plot, it's just a straightforward,
    # common way to get an frb / image array
    prj = ProjectionPlot(
        ds,
        projaxis,
        ("gas", "density"),
        width=(2.5, "cm"),
        weight_field=toweight_field,
        buff_size=(5, 5),
        center=center,
        data_source=source,
        north_vector=northvector,
        depth=depth,
    )
    img = prj.frb.data[("gas", "density")]
    if weighted:
        # periodic shifts will modify the (relative) dl values a bit
        expected_out = np.zeros(
            (
                5,
                5,
            ),
            dtype=img.v.dtype,
        )
        expected_out[::2, ::2] = unitrho
        if depth is None:
            expected_out[2, 2] *= 1.5
        else:
            # only 2 * unitrho element included
            expected_out[2, 2] *= 2.0
    else:
        expected_out = np.zeros(
            (
                5,
                5,
            ),
            dtype=img.v.dtype,
        )
        expected_out[::2, ::2] = dl_cen * prefactor * unitrho
        if depth is None:
            # 3 particles per l.o.s., including the denser one
            expected_out *= 3.0
            expected_out[2, 2] *= 4.0 / 3.0
        else:
            # 1 particle per l.o.s., including the denser one
            expected_out[2, 2] *= 2.0
    # grid is shifted to the left -> 'missing' stuff at the left
    if (not periodic) and shiftcenter:
        expected_out[:1, :] = 0.0
        expected_out[:, :1] = 0.0
    # print(axis, shiftcenter, depth, periodic, weighted)
    # print("expected:\n", expected_out)
    # print("recovered:\n", img.v)
    assert_rel_equal(expected_out, img.v, 4)


_diag_dist = np.sqrt(3.0)  # diagonal distance of a cube with length 1.
# each case is depth, center, expected integrated distance
_cases_to_test = [
    (_diag_dist / 3.0, "domain_left_edge", _diag_dist / 3.0 / 2.0),
    (_diag_dist * 2.0, "domain_left_edge", _diag_dist),
    (_diag_dist * 4.0, "domain_left_edge", _diag_dist),
    (None, "domain_center", _diag_dist),
]


@pytest.mark.parametrize("depth,proj_center,expected", _cases_to_test)
def test_offaxisprojection_depth(depth, proj_center, expected):
    # this checks that the depth keyword argument works as expected.
    # in all cases, it integrates the (index, ones) field for a normal
    # pointing to the right edge corner of the domain.
    #
    # For the tests where the projection is centered on the left edge,
    # the integrate distance will scale as depth / 2.0. When centered
    # on the origin, it will scale with depth. The integrated distance
    # should max out at the diagonal distance of the cube (when the depth
    # exceeds the cube diagonal distance).
    #
    # Also note that the accuracy will depend on the buffer dimensions:
    # using the default (800,800) results in accuracy of about 1 percent

    ds = fake_amr_ds()

    n = [1.0, 1.0, 1.0]
    c = getattr(ds, proj_center)
    field = ("index", "ones")

    p = ProjectionPlot(ds, n, field, depth=depth, weight_field=None, center=c)

    maxval = p.frb[field].max().d
    assert_allclose(expected, maxval, atol=1e-2)


_sph_test_cases = [
    ([1.0, 1.0, 0.7], 27),
    ([1.0, 1.0, 1], 19),
    ([0.0, 0.0, 1], 9),
]


@requires_module_pytest("contourpy")
@pytest.mark.parametrize("normal_vec, n_particles", _sph_test_cases)
def test_offaxisprojection_sph_defaultdepth(normal_vec, n_particles):
    # checks that particles are picked up as expected for a range of
    # depths and normal vectors. Certain viewing angles will result
    # in overlapping particles (since fake_sph_grid_ds aligns particles
    # on a grid): n_particles is the expected number of circles in the
    # resulting image for the given normal vector. Circle counts are
    # calculated here using a contour generator.
    from contourpy import contour_generator

    ds = fake_sph_grid_ds()
    c = ds.domain_center
    diag_dist = np.linalg.norm(ds.domain_width)
    field = ("gas", "mass")
    p = OffAxisProjectionPlot(
        ds, normal_vec, field, weight_field=None, center=c, width=diag_dist
    )
    p.render()

    # get the number of circles in the plot
    cg = contour_generator(z=p.frb[("gas", "mass")].d)
    assert n_particles == len(cg.lines(1.0))



import yt

fn = "IsolatedGalaxy/galaxy0030/galaxy0030"  # dataset to load

ds = yt.load(fn)  # load data

# Now we need a center of our volume to render.  Here we'll just use
# 0.5,0.5,0.5, because volume renderings are not periodic.
c = [0.5, 0.5, 0.5]

# Our image plane will be normal to some vector.  For things like collapsing
# objects, you could set it the way you would a cutting plane -- but for this
# dataset, we'll just choose an off-axis value at random.  This gets normalized
# automatically.
L = [0.5, 0.4, 0.7]

# Our "width" is the width of the image plane as well as the depth.
# The first element is the left to right width, the second is the
# top-bottom width, and the last element is the back-to-front width
# (all in code units)
W = [0.04, 0.04, 0.4]

# The number of pixels along one side of the image.
# The final image will have Npixel^2 pixels.
Npixels = 512

# Now we call the off_axis_projection function, which handles the rest.
# Note that we set no_ghost equal to False, so that we *do* include ghost
# zones in our data.  This takes longer to calculate, but the results look
# much cleaner than when you ignore the ghost zones.
# Also note that we set the field which we want to project as "density", but
# really we could use any arbitrary field like "temperature", "metallicity"
# or whatever.
image = yt.off_axis_projection(ds, c, L, W, Npixels, ("gas", "density"), no_ghost=False)

# Image is now an NxN array representing the intensities of the various pixels.
# And now, we call our direct image saver.  We save the log of the result.
yt.write_projection(
    image,
    "offaxis_projection_colorbar.png",
    colorbar_label="Column Density (cm$^{-2}$)",
)



import os
from contextlib import contextmanager
from pathlib import Path
from shutil import copytree
from tempfile import TemporaryDirectory

import numpy as np
from numpy.testing import assert_equal, assert_raises

import yt
from yt.config import ytcfg
from yt.fields.field_detector import FieldDetector
from yt.frontends.ramses.api import RAMSESDataset
from yt.frontends.ramses.field_handlers import DETECTED_FIELDS, HydroFieldFileHandler
from yt.testing import requires_file, requires_module, units_override_check
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    PixelizedProjectionValuesTest,
    create_obj,
    data_dir_load,
    requires_ds,
)
from yt.utilities.on_demand_imports import _f90nml as f90nml

_fields = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("gas", "velocity_magnitude"),
    ("gas", "pressure_gradient_magnitude"),
)

output_00080 = "output_00080/info_00080.txt"


@requires_ds(output_00080)
def test_output_00080():
    ds = data_dir_load(output_00080)
    assert_equal(str(ds), "info_00080")
    assert_equal(ds.particle_type_counts, {"io": 1090895, "nbody": 0})
    dso = [None, ("sphere", ("max", (0.1, "unitary")))]
    for dobj_name in dso:
        for field in _fields:
            for axis in [0, 1, 2]:
                for weight_field in [None, ("gas", "density")]:
                    yield PixelizedProjectionValuesTest(
                        output_00080, axis, field, weight_field, dobj_name
                    )
            yield FieldValuesTest(output_00080, field, dobj_name)
        dobj = create_obj(ds, dobj_name)
        s1 = dobj["index", "ones"].sum()
        s2 = sum(mask.sum() for block, mask in dobj.blocks)
        assert_equal(s1, s2)


@requires_file(output_00080)
def test_RAMSESDataset():
    assert isinstance(data_dir_load(output_00080), RAMSESDataset)


@requires_file(output_00080)
def test_RAMSES_alternative_load():
    # Test that we can load a RAMSES dataset by giving it the folder name,
    # the info file name or an amr file name
    base_dir, info_file_fname = os.path.split(output_00080)
    for fname in (base_dir, os.path.join(base_dir, "amr_00080.out00001")):
        assert isinstance(data_dir_load(fname), RAMSESDataset)


@requires_file(output_00080)
def test_units_override():
    units_override_check(output_00080)


ramsesNonCosmo = "DICEGalaxyDisk_nonCosmological/output_00002/info_00002.txt"


@requires_file(ramsesNonCosmo)
def test_non_cosmo_detection():
    ds = yt.load(ramsesNonCosmo, cosmological=False)
    assert_equal(ds.cosmological_simulation, 0)

    ds = yt.load(ramsesNonCosmo, cosmological=None)
    assert_equal(ds.cosmological_simulation, 0)

    ds = yt.load(ramsesNonCosmo)
    assert_equal(ds.cosmological_simulation, 0)


@requires_file(ramsesNonCosmo)
def test_unit_non_cosmo():
    for force_cosmo in [False, None]:
        ds = yt.load(ramsesNonCosmo, cosmological=force_cosmo)

        expected_raw_time = 0.0299468077820411  # in ramses unit
        assert_equal(ds.current_time.value, expected_raw_time)

        expected_time = 14087886140997.336  # in seconds
        assert_equal(ds.current_time.in_units("s").value, expected_time)


ramsesCosmo = "output_00080/info_00080.txt"


@requires_file(ramsesCosmo)
def test_cosmo_detection():
    ds = yt.load(ramsesCosmo, cosmological=True)
    assert_equal(ds.cosmological_simulation, 1)

    ds = yt.load(ramsesCosmo, cosmological=None)
    assert_equal(ds.cosmological_simulation, 1)

    ds = yt.load(ramsesCosmo)
    assert_equal(ds.cosmological_simulation, 1)


@requires_file(ramsesCosmo)
def test_unit_cosmo():
    for force_cosmo in [True, None]:
        ds = yt.load(ramsesCosmo, cosmological=force_cosmo)

        # NOTE: these are the old test values, which used 3.08e24 as
        # the Mpc to cm conversion factor
        # expected_raw_time = 1.119216564055017 # in ramses unit
        # expected_time = 3.756241729312462e17 # in seconds

        expected_raw_time = 1.1213297131656712  # in ramses unit
        np.testing.assert_equal(
            ds.current_time.to("code_time").value, expected_raw_time
        )

        expected_time = 3.7633337427123904e17  # in seconds
        np.testing.assert_equal(ds.current_time.to("s").value, expected_time)


ramsesExtraFieldsSmall = "ramses_extra_fields_small/output_00001"


@requires_file(ramsesExtraFieldsSmall)
def test_extra_fields():
    extra_fields = [("particle_family", "I"), ("particle_pointer", "I")]
    ds = yt.load(
        os.path.join(ramsesExtraFieldsSmall, "info_00001.txt"),
        extra_particle_fields=extra_fields,
    )

    # the dataset should contain the fields
    for field, _ in extra_fields:
        assert ("all", field) in ds.field_list

    # Check the family (they should equal 100, for tracer particles)
    dd = ds.all_data()
    families = dd["all", "particle_family"]
    assert all(families == 100)


@requires_file(ramsesExtraFieldsSmall)
def test_extra_fields_2():
    extra_fields = [f"particle_extra_field_{i + 1}" for i in range(2)]
    ds = yt.load(os.path.join(ramsesExtraFieldsSmall, "info_00001.txt"))

    # When migrating to pytest, uncomment this
    # with pytest.warns(
    #     UserWarning,
    #     match=r"Field (.*) has a length \d+, but expected a length of \d+.()",
    # ):
    #     ds.index

    # the dataset should contain the fields
    for field in extra_fields:
        assert ("io", field) in ds.field_list

    # In the dataset, the fields are integers, so we cannot test
    # that they are accessed correctly.


ramses_rt = "ramses_rt_00088/output_00088/info_00088.txt"


@requires_file(ramses_rt)
def test_ramses_rt():
    ds = yt.load(ramses_rt)
    ad = ds.all_data()

    expected_fields = [
        "Density",
        "x-velocity",
        "y-velocity",
        "z-velocity",
        "Pres_IR",
        "Pressure",
        "Metallicity",
        "HII",
        "HeII",
        "HeIII",
    ]

    for field in expected_fields:
        assert ("ramses", field) in ds.field_list

        # test that field access works
        ad["ramses", field]

    # test that special derived fields for RT datasets work
    special_fields = [("gas", "temp_IR")]
    species = ["H_p1", "He_p1", "He_p2"]
    for specie in species:
        special_fields.extend(
            [
                ("gas", f"{specie}_fraction"),
                ("gas", f"{specie}_density"),
                ("gas", f"{specie}_mass"),
            ]
        )

    for field in special_fields:
        assert field in ds.derived_field_list
        ad[field]

    # Test the creation of rt fields
    rt_fields = [("rt", "photon_density_1")] + [
        ("rt", f"photon_flux_{d}_1") for d in "xyz"
    ]
    for field in rt_fields:
        assert field in ds.derived_field_list
        ad[field]


ramses_sink = "ramses_sink_00016/output_00016/info_00016.txt"


@requires_file(ramses_sink)
def test_ramses_sink():
    expected_fields = [
        "BH_bondi_accretion",
        "BH_eddington_accretion",
        "BH_efficiency",
        "BH_esave",
        "BH_real_accretion",
        "BH_spin",
        "BH_spin_x",
        "BH_spin_y",
        "BH_spin_z",
        "gas_spin_x",
        "gas_spin_y",
        "gas_spin_z",
        "particle_birth_time",
        "particle_identifier",
        "particle_mass",
        "particle_position_x",
        "particle_position_y",
        "particle_position_z",
        "particle_prop_0_0",
        "particle_prop_0_1",
        "particle_prop_0_2",
        "particle_prop_0_3",
        "particle_prop_0_4",
        "particle_prop_0_5",
        "particle_prop_0_6",
        "particle_velocity_x",
        "particle_velocity_y",
        "particle_velocity_z",
    ]

    # Check that sinks are autodetected
    ds = yt.load(ramses_sink)
    ad = ds.all_data()

    for field in expected_fields:
        assert ("sink", field) in ds.field_list

        # test that field access works
        ad["sink", field]

    # Checking that sinks are autodetected
    ds = yt.load(ramsesCosmo)
    ad = ds.all_data()

    for field in expected_fields:
        assert ("sink", field) not in ds.field_list


ramses_new_format = "ramses_new_format/output_00002/info_00002.txt"


@requires_file(ramses_new_format)
def test_new_format():
    expected_particle_fields = [
        ("star", "particle_identity"),
        ("star", "particle_level"),
        ("star", "particle_mass"),
        ("star", "particle_metallicity"),
        ("star", "particle_position_x"),
        ("star", "particle_position_y"),
        ("star", "particle_position_z"),
        ("star", "particle_tag"),
        ("star", "particle_velocity_x"),
        ("star", "particle_velocity_y"),
        ("star", "particle_velocity_z"),
    ]

    ds = yt.load(ramses_new_format)
    ad = ds.all_data()

    # Check all the expected fields exist and can be accessed
    for f in expected_particle_fields:
        assert f in ds.derived_field_list
        ad[f]

    # Check there is only stars with tag 0 (it should be right)
    assert all(ad["star", "particle_family"] == 2)
    assert all(ad["star", "particle_tag"] == 0)
    assert len(ad["star", "particle_tag"]) == 600


@requires_file(ramses_sink)
def test_ramses_part_count():
    ds = yt.load(ramses_sink)
    pcount = ds.particle_type_counts

    assert_equal(pcount["io"], 17132, err_msg="Got wrong number of io particle")
    assert_equal(pcount["sink"], 8, err_msg="Got wrong number of sink particle")


@requires_file(ramsesCosmo)
def test_custom_particle_def():
    ytcfg.add_section("ramses-particles")
    ytcfg["ramses-particles", "fields"] = """particle_position_x, d
         particle_position_y, d
         particle_position_z, d
         particle_velocity_x, d
         particle_velocity_y, d
         particle_velocity_z, d
         particle_mass, d
         particle_identifier, i
         particle_refinement_level, I
         particle_birth_time, d
         particle_foobar, d
    """
    ds = yt.load(ramsesCosmo)

    def check_unit(array, unit):
        assert str(array.in_cgs().units) == unit

    try:
        assert ("io", "particle_birth_time") in ds.derived_field_list
        assert ("io", "particle_foobar") in ds.derived_field_list

        check_unit(ds.r["io", "particle_birth_time"], "s")
        check_unit(ds.r["io", "particle_foobar"], "dimensionless")
    finally:
        ytcfg.remove_section("ramses-particles")


@requires_file(output_00080)
@requires_file(ramses_sink)
def test_grav_detection():
    for path, has_potential in ((output_00080, False), (ramses_sink, True)):
        ds = yt.load(path)

        # Test detection
        for k in "xyz":
            assert ("gravity", f"{k}-acceleration") in ds.field_list
            assert ("gas", f"acceleration_{k}") in ds.derived_field_list

        if has_potential:
            assert ("gravity", "Potential") in ds.field_list
            assert ("gas", "potential") in ds.derived_field_list
            assert ("gas", "potential_energy") in ds.derived_field_list

        # Test access
        for k in "xyz":
            ds.r["gas", f"acceleration_{k}"].to("m/s**2")

        if has_potential:
            ds.r["gas", "potential"].to("m**2/s**2")
            ds.r["gas", "potential_energy"].to("kg*m**2/s**2")


@requires_file(ramses_sink)
@requires_file(output_00080)
def test_ramses_field_detection():
    ds1 = yt.load(ramses_rt)

    assert "ramses" not in DETECTED_FIELDS

    # Now they are detected !
    ds1.index
    P1 = HydroFieldFileHandler.parameters
    assert DETECTED_FIELDS[ds1.unique_identifier]["ramses"] is not None
    fields_1 = set(DETECTED_FIELDS[ds1.unique_identifier]["ramses"])

    # Check the right number of variables has been loaded
    assert P1[ds1.unique_identifier]["nvar"] == 10
    assert len(fields_1) == P1[ds1.unique_identifier]["nvar"]

    # Now load another dataset
    ds2 = yt.load(output_00080)
    ds2.index
    P2 = HydroFieldFileHandler.parameters
    fields_2 = set(DETECTED_FIELDS[ds2.unique_identifier]["ramses"])

    # Check the right number of variables has been loaded
    assert P2[ds2.unique_identifier]["nvar"] == 6
    assert len(fields_2) == P2[ds2.unique_identifier]["nvar"]


@requires_file(ramses_new_format)
@requires_file(ramsesCosmo)
@requires_file(ramsesNonCosmo)
def test_formation_time():
    extra_particle_fields = [
        ("particle_birth_time", "d"),
        ("particle_metallicity", "d"),
    ]

    # test semantics for cosmological dataset
    ds = yt.load(ramsesCosmo, extra_particle_fields=extra_particle_fields)
    assert ("io", "particle_birth_time") in ds.field_list
    assert ("io", "conformal_birth_time") in ds.field_list
    assert ("io", "particle_metallicity") in ds.field_list

    ad = ds.all_data()
    whstars = ad["io", "conformal_birth_time"] != 0
    assert np.all(ad["io", "star_age"][whstars] > 0)

    # test semantics for non-cosmological new-style output format
    ds = yt.load(ramses_new_format)
    ad = ds.all_data()
    assert ("io", "particle_birth_time") in ds.field_list
    assert np.all(ad["io", "particle_birth_time"] > 0)

    # test semantics for non-cosmological old-style output format
    ds = yt.load(ramsesNonCosmo, extra_particle_fields=extra_particle_fields)
    ad = ds.all_data()
    assert ("io", "particle_birth_time") in ds.field_list
    # the dataset only includes particles with arbitrarily old ages
    # and particles that formed in the very first timestep
    assert np.all(ad["io", "particle_birth_time"] <= 0)
    whdynstars = ad["io", "particle_birth_time"] == 0
    assert np.all(ad["io", "star_age"][whdynstars] == ds.current_time)


@requires_file(ramses_new_format)
def test_cooling_fields():
    # Test the field is being loaded correctly
    ds = yt.load(ramses_new_format)

    # Derived cooling fields
    assert ("gas", "cooling_net") in ds.derived_field_list
    assert ("gas", "cooling_total") in ds.derived_field_list
    assert ("gas", "heating_total") in ds.derived_field_list
    assert ("gas", "number_density") in ds.derived_field_list

    # Original cooling fields
    assert ("gas", "cooling_primordial") in ds.derived_field_list
    assert ("gas", "cooling_compton") in ds.derived_field_list
    assert ("gas", "cooling_metal") in ds.derived_field_list
    assert ("gas", "heating_primordial") in ds.derived_field_list
    assert ("gas", "heating_compton") in ds.derived_field_list
    assert ("gas", "cooling_primordial_prime") in ds.derived_field_list
    assert ("gas", "cooling_compton_prime") in ds.derived_field_list
    assert ("gas", "cooling_metal_prime") in ds.derived_field_list
    assert ("gas", "heating_primordial_prime") in ds.derived_field_list
    assert ("gas", "heating_compton_prime") in ds.derived_field_list
    assert ("gas", "mu") in ds.derived_field_list

    # Abundances
    assert ("gas", "Electron_number_density") in ds.derived_field_list
    assert ("gas", "HI_number_density") in ds.derived_field_list
    assert ("gas", "HII_number_density") in ds.derived_field_list
    assert ("gas", "HeI_number_density") in ds.derived_field_list
    assert ("gas", "HeII_number_density") in ds.derived_field_list
    assert ("gas", "HeIII_number_density") in ds.derived_field_list

    def check_unit(array, unit):
        assert str(array.in_cgs().units) == unit

    check_unit(ds.r["gas", "cooling_total"], "cm**5*g/s**3")
    check_unit(ds.r["gas", "cooling_primordial_prime"], "cm**5*g/(K*s**3)")
    check_unit(ds.r["gas", "number_density"], "cm**(-3)")
    check_unit(ds.r["gas", "mu"], "dimensionless")
    check_unit(ds.r["gas", "Electron_number_density"], "cm**(-3)")


@requires_file(ramses_rt)
def test_ramses_mixed_files():
    # Test that one can use derived fields that depend on different
    # files (here hydro and rt files)
    ds = yt.load(ramses_rt)

    def _mixed_field(field, data):
        return data["rt", "photon_density_1"] / data["gas", "H_nuclei_density"]

    ds.add_field(("gas", "mixed_files"), function=_mixed_field, sampling_type="cell")

    # Access the field
    ds.r["gas", "mixed_files"]


ramses_empty_record = "ramses_empty_record/output_00003/info_00003.txt"


@requires_file(ramses_empty_record)
def test_ramses_empty_record():
    # Test that yt can load datasets with empty records
    ds = yt.load(ramses_empty_record)

    # This should not fail
    ds.index

    # Access some field
    ds.r["gas", "density"]


@requires_file(ramses_new_format)
@requires_module("f90nml")
def test_namelist_reading():
    ds = data_dir_load(ramses_new_format)
    namelist_fname = os.path.join(ds.directory, "namelist.txt")
    with open(namelist_fname) as f:
        ref = f90nml.read(f)

    nml = ds.parameters["namelist"]

    assert nml == ref


@requires_file(ramses_empty_record)
@requires_file(output_00080)
@requires_module("f90nml")
def test_namelist_reading_should_not_fail():
    for ds_name in (ramses_empty_record, output_00080):
        # Test that the reading does not fail for malformed namelist.txt files
        ds = data_dir_load(ds_name)
        ds.index  # should work


ramses_mhd_128 = "ramses_mhd_128/output_00027/info_00027.txt"


@requires_file(ramses_mhd_128)
def test_magnetic_field_aliasing():
    # Test if RAMSES magnetic fields are correctly aliased to yt magnetic fields and if
    # derived magnetic quantities are calculated
    ds = data_dir_load(ramses_mhd_128)
    ad = ds.all_data()

    for field in [
        "magnetic_field_x",
        "magnetic_field_magnitude",
        "alfven_speed",
        "magnetic_field_divergence",
    ]:
        assert ("gas", field) in ds.derived_field_list
        ad["gas", field]


@requires_file(output_00080)
def test_field_accession():
    ds = yt.load(output_00080)
    fields = [
        ("gas", "density"),  # basic ones
        ("gas", "pressure"),
        ("gas", "pressure_gradient_magnitude"),  # requires ghost zones
    ]
    # Check accessing gradient works for a variety of spatial domains
    for reg in (
        ds.all_data(),
        ds.sphere([0.1] * 3, 0.01),
        ds.sphere([0.5] * 3, 0.05),
        ds.box([0.1] * 3, [0.2] * 3),
    ):
        for field in fields:
            reg[field]


@requires_file(output_00080)
def test_ghost_zones():
    ds = yt.load(output_00080)

    def gen_dummy(ngz):
        def dummy(field, data):
            if not isinstance(data, FieldDetector):
                shape = data["gas", "mach_number"].shape[:3]
                np.testing.assert_equal(shape, (2 + 2 * ngz, 2 + 2 * ngz, 2 + 2 * ngz))
            return data["gas", "mach_number"]

        return dummy

    fields = []
    for ngz in (1, 2, 3):
        fname = ("gas", f"density_ghost_zone_{ngz}")
        ds.add_field(
            fname,
            gen_dummy(ngz),
            sampling_type="cell",
            units="",
            validators=[yt.ValidateSpatial(ghost_zones=ngz)],
        )
        fields.append(fname)

    box = ds.box([0, 0, 0], [0.1, 0.1, 0.1])
    for f in fields:
        print("Getting ", f)
        box[f]


@requires_file(output_00080)
def test_max_level():
    ds = yt.load(output_00080)

    assert any(ds.r["index", "grid_level"] > 2)

    # Should work
    for ds in (
        yt.load(output_00080, max_level=2, max_level_convention="yt"),
        yt.load(output_00080, max_level=8, max_level_convention="ramses"),
    ):
        assert all(ds.r["index", "grid_level"] <= 2)
        assert any(ds.r["index", "grid_level"] == 2)


@requires_file(output_00080)
def test_invalid_max_level():
    invalid_value_args = (
        (1, None),
        (1, "foo"),
        (1, "bar"),
        (-1, "yt"),
    )
    for lvl, convention in invalid_value_args:
        with assert_raises(ValueError):
            yt.load(output_00080, max_level=lvl, max_level_convention=convention)

    invalid_type_args = (
        (1.0, "yt"),  # not an int
        ("invalid", "yt"),
    )
    # Should fail with value errors
    for lvl, convention in invalid_type_args:
        with assert_raises(TypeError):
            yt.load(output_00080, max_level=lvl, max_level_convention=convention)


@requires_file(ramses_new_format)
def test_print_stats():
    ds = yt.load(ramses_new_format)

    # Should work
    ds.print_stats()

    # FIXME #3197: use `capsys` with pytest to make sure the print_stats function works as intended


@requires_file(output_00080)
def test_reading_order():
    # This checks the bug unvovered in #4880
    # This checks that the result of field accession doesn't
    # depend on the order

    def _dummy_field(field, data):
        # Note: this is a dummy field
        # that doesn't really have any physical meaning
        # but may trigger some bug in the field
        # handling.
        T = data["gas", "temperature"]
        Z = data["gas", "metallicity"]
        return T * 1**Z

    fields = [
        "Density",
        "x-velocity",
        "y-velocity",
        "z-velocity",
        "Pressure",
        "Metallicity",
    ]
    ds = yt.load(output_00080, fields=fields)

    ds.add_field(
        ("gas", "test"), function=_dummy_field, units=None, sampling_type="cell"
    )

    ad = ds.all_data()
    v0 = ad["gas", "test"]

    ad = ds.all_data()
    ad["gas", "temperature"]
    v1 = ad["gas", "test"]

    np.testing.assert_allclose(v0, v1)


# Simple context manager that overrides the value of
# the self_shielding flag in the namelist.txt file
@contextmanager
def override_self_shielding(root: Path, section: str, val: bool):
    # Copy content of root in a temporary folder
    with TemporaryDirectory() as tmp:
        tmpdir = Path(tmp) / root.name
        tmpdir.mkdir()

        # Copy content of `root` in `tmpdir` recursively
        copytree(root, tmpdir, dirs_exist_ok=True)

        fname = Path(tmpdir) / "namelist.txt"

        with open(fname) as f:
            namelist_data = f90nml.read(f).todict()
            sec = namelist_data.get(section, {})
            sec["self_shielding"] = val
            namelist_data[section] = sec

        with open(fname, "w") as f:
            new_nml = f90nml.Namelist(namelist_data)
            new_nml.write(f)

        yield tmpdir


@requires_file(ramses_new_format)
@requires_module("f90nml")
def test_self_shielding_logic():
    ##################################################
    # Check that the self_shielding flag is correctly read
    ds = yt.load(ramses_new_format)

    assert ds.parameters["namelist"] is not None
    assert ds.self_shielding is False

    ##################################################
    # Modify the namelist in-situ, reload and check
    root_folder = Path(ds.root_folder)
    for section in ("physics_params", "cooling_params"):
        for val in (True, False):
            with override_self_shielding(root_folder, section, val) as tmp_output:
                # Autodetection should work
                ds = yt.load(tmp_output)
                assert ds.parameters["namelist"] is not None
                assert ds.self_shielding is val

                # Manually set should ignore the namelist
                ds = yt.load(tmp_output, self_shielding=True)
                assert ds.self_shielding is True

                ds = yt.load(tmp_output, self_shielding=False)
                assert ds.self_shielding is False

    ##################################################
    # Manually set should work, even if namelist does not contain the flag
    ds = yt.load(ramses_new_format, self_shielding=True)
    assert ds.self_shielding is True

    ds = yt.load(ramses_new_format, self_shielding=False)
    assert ds.self_shielding is False


ramses_star_formation = "ramses_star_formation/output_00013"


@requires_file(ramses_star_formation)
@requires_module("f90nml")
def test_self_shielding_loading():
    ds_off = yt.load(ramses_star_formation, self_shielding=False)
    ds_on = yt.load(ramses_star_formation, self_shielding=True)

    # We do not expect any significant change at densities << 1e-2
    reg_on = ds_on.all_data().cut_region(
        ["obj['gas', 'density'].to('mp/cm**3') < 1e-6"]
    )
    reg_off = ds_off.all_data().cut_region(
        ["obj['gas', 'density'].to('mp/cm**3') < 1e-6"]
    )

    np.testing.assert_allclose(
        reg_on["gas", "cooling_total"],
        reg_off["gas", "cooling_total"],
        rtol=1e-5,
    )

    # We expect large differences from 1e-2 mp/cc
    reg_on = ds_on.all_data().cut_region(
        ["obj['gas', 'density'].to('mp/cm**3') > 1e-2"]
    )
    reg_off = ds_off.all_data().cut_region(
        ["obj['gas', 'density'].to('mp/cm**3') > 1e-2"]
    )
    # Primordial cooling decreases with density (except at really high temperature)
    # so we expect the boosted version to cool *less* efficiently
    diff = (
        reg_on["gas", "cooling_primordial"] - reg_off["gas", "cooling_primordial"]
    ) / reg_off["gas", "cooling_primordial"]
    assert (np.isclose(diff, 0, atol=1e-5) | (diff < 0)).all()

    # Also make sure the difference is large for some cells
    assert (np.abs(diff) > 0.1).any()


@requires_file(output_00080)
@requires_file(ramses_mhd_128)
def test_order_does_not_matter():
    for order in (1, 2):
        ds0 = yt.load(output_00080)
        ds1 = yt.load(ramses_mhd_128)

        # This should not raise any exception
        if order == 1:
            _sp1 = ds1.all_data()
            sp0 = ds0.all_data()
        else:
            sp0 = ds0.all_data()
            _sp1 = ds1.all_data()

        sp0["gas", "velocity_x"].max().to("km/s")
